import uuid
import maya

from typing import Optional, List
from jamboree import Jamboree
from jamboree.base.processors.abstracts import Processor
from jamboree.handlers.abstracted.search import ParameterizedSearch
from jamboree.handlers.default.search import BaseSearchHandler
from see137.search.prototype import SearchFacade
from loguru import logger

class _BacktestStrategyManagement(ParameterizedSearch):
    """ 
        Use to find backtest information by episodes, and episodes by backtest

        - Is there any backtest that's running right now?
        - Is the episode still running?
        - Update the episode's status
        - Start an episode
        - Get the number of episodes that are running

        ```py
            # Start a new backtest
            backtest_manager.create.episodes.init(backtest_id)

            backtest_manager.update.episodes.status(backtest_id, "...")

            # Get the difference
            backtest_manager.info.episodes.by_backtest(backtest_id)
            backtest_manager.info.episodes.by_episode(episode_id)

            
            # General running episodes
            backtest_manager.info.episodes.total
            backtest_manager.info.episodes.total_active
            backtest_manager.info.episodes.total_inactive



            backtest_manager.info.episodes.active_by(backtest)
            backtest_manager.info.episodes.inactive_by(backtest)
            backtest_manager.info.episodes.is_running(backtest_id, by="backtest")
            backtest_manager.info.episodes.is_running(episode_id, by="episode")

            backtest_manager.info.episodes.status(episode_id, by="episode")
            backtest_manager.info.episodes.status(backtest_id, by="backtest")
        ```
    """
    def __init__(self):
        super().__init__()
        self.entity = "stategy_online"
        self.dreq = {"strategy": str, "broker": str, "user":str, "live": bool}
        

class BacktestStrategyManagement(SearchFacade):
    def __init__(self):
        super().__init__(_BacktestStrategyManagement())
        self.must_have = ['strategy', 'userid', 'broker']
        self.allowed = ["RUNNING", "DONE"]

    # Universal functions that we'll use in other parts

    def active_by(self, backtest_id:str) -> List[dict]:
        """ Get all active backtests by backtest_id"""
        return self.Find(backtest=backtest_id, status="RUNNING")

    def inactive_by(self, backtest_id:str) -> List[dict]:
        """ Get all inactive backtests by backtest_id"""
        return self.Find(backtest=backtest_id, status="DONE")
    
    def total_active_by(self, backtest_id:str) -> int:
        return len(self.active_by(backtest_id))

    def total_inactive_by(self, backtest_id:str) -> int:
        return len(self.inactive_by(backtest_id))


class StrategyCreate(BacktestStrategyManagement):
    """ 
        Use to find backtest information by episodes, and episodes by backtest

        - Is there any backtest that's running right now?
        - Is the episode still running?
        - Update the episode's status
        - Start an episode
        - Get the number of episodes that are running

        ```py
            # Start a new backtest
            backtest_manager.create.episodes.init(backtest_id)
        ```
    """
    def __init__(self):
        super().__init__()
    
    def init(self, strategy_metadata_id:str, userid:str, broker:str, live:bool=False):
        strategy_id = self.Create(no_overwrite_reqs=True, strategy=strategy_metadata_id, broker=broker, live=live, userid=userid)
        return strategy_id
    
    
class StrategyUpdate(BacktestStrategyManagement):
    """ 
        Use to find backtest information by episodes, and episodes by backtest

        - Is there any backtest that's running right now?

        ```py
            # Start a new backtest

            backtest_manager.update.strategy.status(strategyid, "LIVE")
        ```
    """
    def __init__(self):
        super().__init__()
        
    
    def status(self, episode_id:str, status="RUNNING"):
        """ Update episode status """
        self.UpdateMany({"episode": episode_id}, status="DONE")
        

    


class StrategyInfo(BacktestStrategyManagement):
    """ 
        Use to find backtest information by episodes, and episodes by backtest

        - Is there any backtest that's running right now?
        - Is the episode still running?
        - Update the episode's status
        - Start an episode
        - Get the number of episodes that are running

        ```py
            # Get the difference
            backtest_manager.info.episodes.by_backtest(backtest_id)
            backtest_manager.info.episodes.by_episode(episode_id)

            
            # General running episodes
            backtest_manager.info.episodes.total
            backtest_manager.info.episodes.total_active
            backtest_manager.info.episodes.total_inactive

            # Get active episodes
            # Check if they're running
            backtest_manager.info.episodes.active_by(backtest)
            backtest_manager.info.episodes.inactive_by(backtest)
            backtest_manager.info.episodes.is_running(backtest_id, by="backtest")
            backtest_manager.info.episodes.is_running(episode_id, by="episode")

            backtest_manager.info.episodes.status(episode_id, by="episode")
            backtest_manager.info.episodes.status(backtest_id, by="backtest")
        ```
    """
    def __init__(self):
        super().__init__()
    
    def strategy_by_asset(self, asset_id:str):
        """ Get the strategy by asset """
        assets = self.Find(asset=asset_id)

class StrategyRemove(BacktestStrategyManagement):
    """ 
        
    """
    def __init__(self):
        super().__init__()

    def by_id(self, episode_id:str):
        """ Remove episode by id"""
        self.Remove(episode=episode_id)

    def by_backtest(self, backtest_id:str):
        self.Remove(backtest=backtest_id)


def manage_episodes():
    proc = Jamboree()
    # episode_create = CoreEpisodeCreate()
    # episode_update = CoreEpisodeUpdate()
    # episode_info = CoreEpisodeInfo()
    # episode_remove = CoreEpisodeRemove()


    # episode_create.processor = proc
    # episode_update.processor = proc
    # episode_info.processor = proc
    # episode_remove.processor = proc

    # backtest_id = "3e667a9ade4c4a409ae153522f46cc3d"

    # backtest_id, episode_id, is_running = episode_create.init(backtest_id)
    # print(episode_info.total_active)
    # print(episode_info.status(episode_id))
    # print(episode_info.is_running(episode_id))
    # episode_update.status(episode_update, status="DONE")
    # print(episode_info.status(episode_id))
    # print(episode_info.total_active)
    # print(episode_info.total_inactive)
    # print(episode_info.total_active_by(backtest_id))
    # episode_remove.by_id(episode_id)
    # print(episode_info.total_inactive)

if __name__ == "__main__":
    manage_episodes()


from jamboree import Jamboree, DBHandler # This is how we connect to databases and outside systems
import pandas as pd
import uuid
import time
from jamboree.utils.core import consistent_hash
# Handlers are ways to manage events and everything related to events. 
# They abstract all of the push and pull operations

class OrderHandler(DBHandler):
    def __init__(self):
        # We totally need to call this. Don't forget to do it.
        super().__init__()
        # The code base checks for these fields prior to any insert operations
        # They're also used to generate a key. A key is like a lane where we append events in a sorted manner.
        # This is also useful when you need information for other parts of the system, such as to keep track of time.
        self.entity = "orders" # An entity is a name of a real world distinguishable object
        self.required = {
            "exchange": str,
            "broker": str,
            "user_id": str,
            "episode": str,
            "live": bool
        }
        self._is_event = False
        self.order_fields = ['status', 'price', 'amount']
        self.allowed_statuses = ["OPEN", "PARTIAL", "FILLED", "CANCELLED"]
        self.order_status_paths = {
            "OPEN": ["PARTIAL", "FILLED", "CANCELLED"],
            "PARTIAL": ["FILLED", "CANCELLED"]
        }
        self.transistionable = ["OPEN", "PARTIAL"]
    
    @property
    def history(self):
        """ Returns a list of all the last 500 orders with this specific profile in a dataframe """
        all_orders = self.many(ar="relative")
        if len(all_orders) == 0:
            return []
        order_frame = pd.DataFrame(all_orders)
        _to_datetime = pd.to_datetime(order_frame.time, unit='s')
        order_frame = order_frame.drop(columns=['mtype', 'user_id', "broker", "exchange", "live", "timestamp", "episode", "time"])
        order_frame.set_index(_to_datetime, inplace=True)
        return order_frame
    
    @property
    def open_orders_spec(self):
        """ Get all open orders by exchange, broker, user_id, episode, live """
        pass
    
    @property
    def closed_orders_spec(self):
        """ Get all open orders by exchange, broker, user_id, episode, live """
        pass
    
    @property
    def current_time(self):
        return time.time()

    def is_valid_order(self, _order:dict):
        return all(field in _order.keys() for field in self.order_fields)
    
    def check_order_status(self, status:str, prior:str):
        if prior in self.transistionable:
            allowed_transitions = self.order_status_paths[prior]
            if status in allowed_transitions:
                return True
        return False
    
    def save_order(self, price:float, amount:float, status:str, order_number:int, order_id:str=None):
        """ Saves order information """
        self['order_num'] = order_number
        # Get the last order with that order number
        if status not in self.allowed_statuses:
            return
        """
            `self.last()`
        """ 
        previous_order = self.last()
        if len(previous_order.keys()) == 0:
            if status != "OPEN":
                print("New orders must be 'OPEN'")
                return
            order = {}
            _total = price * amount
            order['price'] = price
            order['amount'] = amount
            order['status'] = status
            order['total'] = _total
            order['time'] = self.current_time
            self.save(order)
            return
        
        prior_stat = previous_order.get('status')
        valid_prior = self.check_order_status(status, prior_stat)
        if valid_prior:
            order = {}
            _total = price * amount
            order['price'] = price
            order['amount'] = amount
            order['status'] = status
            order['total'] = _total
            order['time'] = self.current_time
            self.save(order)
        else:
            print("Not valid prior status")


import datetime as dt
import random
import time
import uuid
from typing import Dict, List

from addict import Dict as ADict
from jamboree import Jamboree
from jamboree.handlers.default.multi import MultiDataManagement
from jamboree.handlers.default.search import BaseSearchHandler
from loguru import logger

from see137 import PortfolioHandler, RequirementsHandler
from see137.common import (AddictPipeline, AddictWithManagerComponent,
                                    ComponentAbstract, Pipeline, PipelineA)
from see137.components.backtest import CreateBacktestComponent
from see137.components.strategy import CheckStrategy
from see137.components.universe import CheckUniverse
from see137.handlers.strategy.core import StrategyEngine
from see137.models.backtests_models import BacktestConfig
from see137.search.prototype import SearchFacade
from see137.search.backtesting.topics.portfolio import BacktestPortfolioAgg

class CreateEpisodePortfolio(AddictWithManagerComponent):
    def __init__(self):
        super(CreateEpisodePortfolio, self).__init__()
        # Backtest configuration dict. Can be saved into the database immediately.
        self.required_items: Dict[str, type] = {
            "config": dict,
            "valid_multi": bool,
            'valid_strategy': bool,
            'backtestid': str
        }

    def add_backtest_info(self, item: ADict, _id: str, is_save: bool = False):
        item.backtest_id = _id
        item.is_saved = False
        return item

    # def check_prev()

    def process(self, item: ADict):
        """Take the backtest_config and save it into the database.

        Arguments:
            item {ADict} -- Addict modified dictionary. Save's according to backtest data.

        Returns:
            ADict -- Modified dictionary with action and percentage inside of it
        """
        _, episodeid, is_created = (self.manager.create.episodes.init(
            item.backtestid, strategy=item.config.strategy))
        logger.debug(item)
        if is_created == True:
            portfolio_agg = BacktestPortfolioAgg()
            portfolio_agg.processor = self.processor
            aggregationid = (portfolio_agg.Create(
                no_overwrite_reqs=True,
                timestamp=item.config.start,
                episode=episodeid,
                user=item.config.user,
                total=item.config.initial,
                allocation={item.config.base: 1.0},
                performance=item.config.initial))
            agg_msg = f"Aggregate portfolio information ID {aggregationid}"
            logger.success(agg_msg)
            # logger.debug(aggregationid)
            self.manager.update.episodes.portfolio(episodeid, aggregationid)
            _data = self.manager.info.episodes.by_episode(episodeid)
            item.current_episode = episodeid
            item.failed = False
            item.msg = "Backtest was started"
            return item
        elif not bool(episodeid):
            item.current_episode = episodeid
            item.failed = True
            item.msg = "Backtest couldn't be found"
            return item
        item.current_episode = episodeid
        item.failed = True
        item.msg = "Episode already started"
        return item




import uuid
from typing import Any, Dict, List

import maya
import pandas as pd
import yfinance as yf
from addict import Dict as ADDict
from anycache import anycache
from jamboree import Jamboree
from jamboree.handlers import DataHandler, MultiDataManagement
from jamboree.handlers.abstracted.datasets.price import PriceData
from loguru import logger


@anycache(cachedir='/tmp/anycache.my')
def get_dataset_info(ds_str: str):
    # ds_str = " ".join(ds_list)
    ticker = yf.Ticker(ds_str)

    return ticker.info


@anycache(cachedir='/tmp/anyticker.my')
def get_ticker_data(ds_str: str):
    data = yf.download(ds_str, group_by='ticker').dropna()
    return data


class CallablePriceData(PriceData):
    def __init__(self):
        super().__init__()
        self.forced_subcategories = ['market', 'country', 'sector', 'exchange']
        self.specs = [
            'market', 'country', 'sector', 'exchange', 'abbreviation', 'name'
        ]

    def __call__(self,
                 name: str,
                 abbv: str,
                 market: str = "crypto",
                 country: str = "US",
                 sector: str = "technology",
                 exchange: str = "coinbase",
                 **kwargs):
        self['name'] = name
        self['abbreviation'] = abbv
        self['subcategories'] = {
            "market": market,
            "country": country,
            "sector": sector,
            "exchange": exchange,
        }

        # TODO: Figure out how to add random search fields.

        # for k, v in kwargs.items():
        #     if k not in self.specs:
        #         self[k] = v
        return self

    def dataframe_from_last(self):
        """ Get a dataframe between a head and tail. Resample according to our settings"""

        values = self.many(ar="relative")
        frame = pd.DataFrame(values)
        frame = self._timestamp_resample_and_drop(frame)
        return frame


class DatasetCommander:
    """ 
        Use this to manage individual datasets. 
        * Create 
    """
    def __init__(self, *args, **kwargs):
        self.processor = Jamboree()
        self.pricedata_management = CallablePriceData()
        self.multi_data = MultiDataManagement()
        self.pricedata_management.processor = self.processor
        self.multi_data.processor = self.processor

    def AllMeta(self) -> List[Dict[str, Any]]:
        """ Get all metadata for datasets """
        return []

    def AllPrice(self) -> List[Dict[str, Any]]:
        """ Get all price metadata out there. """
        _current_search = self.pricedata_management.search
        res = _current_search.find()
        return res

    def AllUniverses(self):
        """
            All Universes

            Get all universes available.
        """
        current_search = self.multi_data.search
        return current_search.find()

    def AttachMetaParams(self, meta_params):
        """ 
            Attach time information and other parameters we need for operations to metadata.

            Use to know when to pull information for a given piece of metadata.

            * Find all metadata IDs that are over current time.
            * Lock current position inside of database
            * Set operation that they're being worked on
            * On pull, update the last time we operated on such data.
        """
        pass

    def RemoveAllPrice(self):
        """Removes all of the price information.

        Finds all of the price information given the supplied information general fields then removes the metadata.

        """
        _current_search = self.pricedata_management.search
        _current_search.remove()

    def GetPriceHandler(self, _id: str):
        """Get Price Handler

        Get a price handler by id. 

        Parameters
        ----------
        _id : str
            metadata id for the price handler. 
        """
        _current_search = self.pricedata_management.search
        full_metadata = _current_search.FindById(_id)
        if not bool(full_metadata):
            return None
        gen = ADDict(**full_metadata)

        self.pricedata_management(name=gen.name,
                                  abbv=gen.abbreviation,
                                  market=gen.subcategories.market,
                                  country=gen.subcategories.country,
                                  sector=gen.subcategories.sector,
                                  exchange=gen.subcategories.exchange)
        self.pricedata_management.reset()
        return self.pricedata_management.copy()
        # print(self.pricedata_management.dataframe_from_last())


def run_dataset_commandments():
    callable_price = CallablePriceData()
    callable_price(name="Blank",
                   abbv="BLK",
                   market="stocks",
                   country="US",
                   sector="technology",
                   exchage="dow jones")
    callable_price.live = False
    callable_price.episode = uuid.uuid4().hex

    # Download assets


if __name__ == "__main__":
    run_dataset_commandments()



import math
import numpy
import random
import numpy.random as nrand
from see137.stochastic import ModelParameters

def cox_ingersoll_ross_heston(param):
    """
    This method returns the rate levels of a mean-reverting cox ingersoll ross process. It is used to model interest
    rates as well as stochastic volatility in the Heston model. Because the returns between the underlying and the
    stochastic volatility should be correlated we pass a correlated Brownian motion process into the method from which
    the interest rate levels are constructed. The other correlated process is used in the Heston model
    :param param: the model parameters objects
    :return: the interest rate levels for the CIR process
    """
    # We don't multiply by sigma here because we do that in heston
    sqrt_delta_sigma = math.sqrt(param.all_delta) * param.all_sigma
    brownian_motion_volatility = nrand.normal(loc=0, scale=sqrt_delta_sigma, size=param.all_time)
    a, mu, zero = param.heston_a, param.heston_mu, param.heston_vol0
    volatilities = [zero]
    for i in range(1, param.all_time):
        drift = a * (mu - volatilities[i-1]) * param.all_delta
        randomness = math.sqrt(volatilities[i - 1]) * brownian_motion_volatility[i - 1]
        volatilities.append(volatilities[i - 1] + drift + randomness)
    return numpy.array(brownian_motion_volatility), numpy.array(volatilities)


def heston_model_levels(param):
    """
    NOTE - this method is dodgy! Need to debug!
    The Heston model is the geometric brownian motion model with stochastic volatility. This stochastic volatility is
    given by the cox ingersoll ross process. Step one on this method is to construct two correlated GBM processes. One
    is used for the underlying asset prices and the other is used for the stochastic volatility levels
    :param param: model parameters object
    :return: the prices for an underlying following a Heston process
    """
    assert isinstance(param, ModelParameters)
    # Get two correlated brownian motion sequences for the volatility parameter and the underlying asset
    # brownian_motion_market, brownian_motion_vol = get_correlated_paths_simple(param)
    brownian, cir_process = cox_ingersoll_ross_heston(param)
    brownian, brownian_motion_market = heston_construct_correlated_path(param, brownian)

    heston_market_price_levels = [param.all_s0]
    for i in range(1, param.all_time):
        drift = param.gbm_mu * heston_market_price_levels[i - 1] * param.all_delta
        vol = cir_process[i - 1] * heston_market_price_levels[i - 1] * brownian_motion_market[i - 1]
        heston_market_price_levels.append(heston_market_price_levels[i - 1] + drift + vol)
    return numpy.array(heston_market_price_levels), numpy.array(cir_process)

def heston_construct_correlated_path(param, brownian_motion_one):
    """
    This method is a simplified version of the Cholesky decomposition method for just two assets. It does not make use
    of matrix algebra and is therefore quite easy to implement.
    :param param: model parameters object
    :return: a correlated brownian motion path
    """
    # We do not multiply by sigma here, we do that in the Heston model
    sqrt_delta = math.sqrt(param.all_delta)
    # Construct a path correlated to the first path
    brownian_motion_two = []
    for i in range(param.all_time - 1):
        term_one = param.cir_rho * brownian_motion_one[i]
        term_two = math.sqrt(1 - math.pow(param.cir_rho, 2.0)) * random.normalvariate(0, sqrt_delta)
        brownian_motion_two.append(term_one + term_two)
    return numpy.array(brownian_motion_one), numpy.array(brownian_motion_two)


import time
import maya
import datetime
import humanize
import uuid

from pprint import pprint
from typing import Dict, Any
from addict import Dict as ADict

from jamboree import Jamboree
from loguru import logger
from see137.common import AddictWithManagerComponent
from see137.search.backtesting import BacktestManager



class RemoveBacktestComponent(AddictWithManagerComponent):
    """Load Backtest 

        Load the configuration information for a given backtest.
    """
    def __init__(self):
        super().__init__()
        self.name = "Remove Backtest"
        # Backtest configuration dict. Can be saved into the database immediately.
        
        self.required_items = {
            "backtest": str
        }



    def process(self, item:ADict):
        """Given an id inside of the item dictionary `backtest`, get information using that id

        Arguments:
            item {ADict} -- Addict modified dictionary. Save's according to backtest data.

        Returns:
            ADict -- Modified dictionary with action and percentage inside of it
        """
        logger.error(item)
        return item


    def step(self, item:ADict, **settings):
        return super().step(item, **settings)

if __name__ == "__main__":
    response = None
    # for _ in range(100):
    sample_backtest_dict = ADict(backtest=uuid.uuid4().hex)

    load_backtest = RemoveBacktestComponent()
    load_backtest.processor = Jamboree()
    response = load_backtest.step(sample_backtest_dict)
    
    
    
    pprint(response)
    



import uuid
import maya

from typing import Optional
from jamboree import Jamboree
from jamboree.base.processors.abstracts import Processor
from jamboree.handlers.default.search import BaseSearchHandler

from loguru import logger

class SearchFacade(object):
    def __init__(self, search_handler:BaseSearchHandler, is_reset=False):
        self.must_have = []
        self.is_reset = is_reset
        
        
        
        # Hidden variables with properties
        self._search:BaseSearchHandler = search_handler
    
    @property
    def processor(self):
        if self._processor is None:
            raise AttributeError("The processor hasn't been set yet.")
        return self._processor

    @processor.setter
    def processor(self, _processor:Processor):
        self._processor = _processor
    
    @property
    def search(self):
        self._search.processor = self.processor
        return self._search
    
    def enable_reset(self):
        self.is_reset = True
    
    def disable_reset(self):
        self.is_reset = False

    def __check_requirements(self, items: dict):
        for _abs in self.must_have:
            if _abs not in items:
                raise NotImplementedError(
                    f"{_abs} has to be added. The absolute required variables are the following: {self.must_have}"
                )

    def __check_other_forced(self, forced: list, items: dict):
        for k in forced:
            if k not in items:
                raise AttributeError(f"{k} not in the arguments")

    def Create(self,
               allow_duplicates=False,
               no_overwrite_reqs=False,
               **kwargs):
        self.search.reset()
        self.__check_requirements(kwargs)

        if no_overwrite_reqs and len(self.must_have) > 0:
            
            _all = self.FForced(**kwargs)
            if len(_all) > 0:
                return _all[0].get("id")

        for k, v in kwargs.items():
            self.search[k] = v

        identity = self.search.insert(allow_duplicates=allow_duplicates)
        return identity

    def UpdateID(self, identity: str, **kwargs):
        """ Updates a record by ID. Gives a warning if you're using a must have variable."""
        self.reset()
        for k, v in kwargs.items():
            self.search.replacement[k] = v
        self.search.update_id(identity)

    def UpdateMany(self, search_dict: dict, force_must=False, **replacements):
        self.reset()
        if not bool(search_dict) or not bool(replacements):
            raise ValueError(
                "You need to have query information AND something to replace it with."
            )
        if force_must:
            self.__check_requirements(search_dict)

        for k, v in search_dict.items():
            self.search[k] = v

        for k, v in replacements.items():
            self.search.replacement[k] = v
        self.search.update()

    def Find(self, general=None,force_must=False, **fields):
        self.reset()
        if general is not None:
            self.general = general
        if not bool(fields):
            if general is not None:
                return self.search.find()
            raise ValueError("You have to search using something")
        if force_must:
            self.__check_requirements(fields)

        for k, v in fields.items():
            self.search[k] = v
        return self.search.find()

    def FindById(self, identity: str):
        self.reset()
        remainder = self.search.pick(identity)
        return remainder
    
    def FindForced(self, **kwargs):
        self.reset()
        self.__check_requirements(kwargs)
        for k in self.must_have:
            self.search[k] = kwargs.get(k)
        _all = self.search.find()
        return _all
    
    def FForced(self, **kwargs):
        for k in self.must_have:
            self.search[k] = kwargs.get(k)
        _all = self.search.find()
        return _all
    

    def Remove(self, **kwargs):
        self.reset()
        for k, v in kwargs.items():
            self.search[k] = v
        self.search.remove()
    
    def RemoveAll(self):
        self.search.reset()
        self.general = "*"
        self.search.remove()
    
    def reset(self):
        if self.is_reset:
            self.search.reset()
    
    def first(self, general=None, force_must=False, **kwargs):
        found = self.Find(general=general, force_must=force_must, **kwargs)
        if len(found) > 0:
            return found[0]
        raise AttributeError("No items matching the given profile was found.")

class PrototypeSearchFeatures(BaseSearchHandler):
    def __init__(self):
        super().__init__()
        self.must_have = []
        # Idea, try using upper case to distinguish from the base application.

    def __check_requirements(self, items: dict):
        for _abs in self.must_have:
            if _abs not in items:
                raise NotImplementedError(
                    f"{_abs} has to be added. The absolute required variables are the following: {self.must_have}"
                )

    def __check_other_forced(self, forced: list, items: dict):
        for k in forced:
            if k not in items:
                raise AttributeError(f"{k} not in the arguments")

    def Create(self,
               allow_duplicates=False,
               no_overwrite_reqs=False,
               **kwargs):
        self.reset()
        self.__check_requirements(kwargs)

        if no_overwrite_reqs and len(self.must_have) > 0:
            
            _all = self.FForced(**kwargs)
            if len(_all) > 0:
                return _all[0].id

        for k, v in kwargs.items():
            self[k] = v

        identity = self.insert(allow_duplicates=allow_duplicates)
        return identity

    def UpdateID(self, identity: str, **kwargs):
        """ Updates a record by ID. Gives a warning if you're using a must have variable."""
        self.reset()
        for k, v in kwargs.items():
            self.replacement[k] = v
        self.update_id(identity)

    def UpdateMany(self, search_dict: dict, force_must=False, **replacements):
        self.reset()
        if not bool(search_dict) or not bool(replacements):
            raise ValueError(
                "You need to have query information AND something to replace it with."
            )
        if force_must:
            self.__check_requirements(search_dict)

        for k, v in search_dict.items():
            self[k] = v

        for k, v in replacements.items():
            self.replacement[k] = v
        self.update()

    def Find(self, general=None,force_must=False, **fields):
        self.reset()
        if general is not None:
            self.general = general
        if not bool(fields):
            if general is not None:
                return self.find()
            raise ValueError("You have to search using something")
        if force_must:
            self.__check_requirements(fields)

        for k, v in fields.items():
            self[k] = v
        return self.find()

    def FindById(self, identity: str):
        self.reset()
        remainder = self.pick(identity)
        return remainder
    
    def FindForced(self, **kwargs):
        self.reset()
        self.__check_requirements(kwargs)
        for k in self.must_have:
            self[k] = kwargs.get(k)
        _all = self.find()
        return _all
    
    def FForced(self, **kwargs):
        for k in self.must_have:
            self[k] = kwargs.get(k)
        _all = self.find()
        return _all

    def Remove(self, **kwargs):
        self.reset()
        for k, v in kwargs.items():
            self[k] = v
        self.remove()
    



import math
import numpy
from see137.stochastic import brownian_motion_log_returns


def cox_ingersoll_ross_levels(param):
    """
        This method returns the rate levels of a mean-reverting cox ingersoll ross process. It is used to model interest
        rates as well as stochastic volatility in the Heston model. Because the returns between the underlying and the
        stochastic volatility should be correlated we pass a correlated Brownian motion process into the method from which
        the interest rate levels are constructed. The other correlated process is used in the Heston model
        :param param: the model parameters object
        :return: the interest rate levels for the CIR process
    """
    brownian_motion = brownian_motion_log_returns(param)
    # Setup the parameters for interest rates
    a, mu, zero = param.cir_a, param.cir_mu, param.all_r0
    # Assumes output is in levels
    levels = [zero]
    for i in range(1, param.all_time):
        drift = a * (mu - levels[i-1]) * param.all_delta
        # The main difference between this and the Ornstein Uhlenbeck model is that we multiply the 'random'
        # component by the square-root of the previous level i.e. the process has level dependent interest rates.
        randomness = math.sqrt(levels[i - 1]) * brownian_motion[i - 1]
        levels.append(levels[i - 1] + drift + randomness)
    return numpy.array(levels)


# Copyright 2019 The TensorTrade Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

from itertools import product
from typing import Union, List, Dict
from collections import OrderedDict




class Broker(object):
    """
        A broker for handling the execution of orders on multiple exchanges.
        Orders are kept in a virtual order book until they are ready to be executed.
    """

    def __init__(self):
        
        self._unexecuted = []
        self._executed = {}
        self._trades = OrderedDict()

    # @property
    # def exchanges(self) -> List['Exchange']:
    #     """The list of exchanges the broker will execute orders on."""
    #     return self._exchanges

    # @exchanges.setter
    # def exchanges(self, exchanges: Union[List['Exchange'], 'Exchange']):
    #     self._exchanges = exchanges if isinstance(exchanges, list) else [exchanges]

    # @property
    # def unexecuted(self) -> List[Order]:
    #     """The list of orders the broker is waiting to execute, when their criteria is satisfied."""
    #     return self._unexecuted

    # @property
    # def executed(self) -> Dict[str, Order]:
    #     """The dictionary of orders the broker has executed since resetting, organized by order id"""
    #     return self._executed

    # @property
    # def trades(self) -> Dict[str, 'Trade']:
    #     """The dictionary of trades the broker has executed since resetting, organized by order id."""
    #     return self._trades

    # def submit(self, order: Order):
    #     self._unexecuted += [order]

    # def cancel(self, order: Order):
    #     if order.status == OrderStatus.CANCELLED:
    #         raise Warning(
    #             'Cannot cancel order {} - order has already been cancelled.'.format(order.id))

    #     if order.status != OrderStatus.PENDING:
    #         raise Warning(
    #             'Cannot cancel order {} - order has already been executed.'.format(order.id))

    #     self._unexecuted.remove(order)

    #     order.cancel()

    # def update(self):
    #     for order, exchange in product(self._unexecuted, self._exchanges):
    #         is_executable = order.is_executable_on(exchange) and self.clock.step >= order.start

    #         if order in self._unexecuted and is_executable:
    #             self._unexecuted.remove(order)
    #             self._executed[order.id] = order

    #             order.attach(self)
    #             order.execute(exchange)

    #     for order in self._unexecuted + list(self._executed.values()):
    #         order_expired = (self.clock.step >= order.end) if order.end else False
    #         order_active = order.status not in [OrderStatus.FILLED, OrderStatus.CANCELLED]

    #         if order_active and order_expired:
    #             order.cancel()

    # def on_fill(self, order: Order, exchange: 'Exchange', trade: 'Trade'):
    #     if trade.order_id in self._executed and trade not in self._trades:
    #         self._trades[trade.order_id] = self._trades.get(trade.order_id, [])
    #         self._trades[trade.order_id] += [trade]

    #         if order.is_complete():
    #             next_order = order.complete(exchange)

    #             if next_order:
    #                 self.submit(next_order)

    # def reset(self):
    #     self._unexecuted = []
    #     self._executed = {}
    #     self._trades = OrderedDict()



class OrderCycle(object):
    """ 
        # Order Cycle

        Use in conjucntion with the rest of the pipeline.

        1. Create orders
        2. Change status of orders
        3. Finish orders
        4. Get history of individual order
    """
    def __init__(self):
        pass
    pass


import uuid
import maya

from typing import Optional
from jamboree import Jamboree
from jamboree.base.processors.abstracts import Processor
# from jamboree.handlers.abstracted.search import ParameterizedSearch


from see137.search.backtesting import CreateAccessor, InfoAccessor, UpdateAccessor, RemoveAccessor
from see137.common.injectable import InjectableObject
from loguru import logger


class BacktestManager(InjectableObject):
    """ 
        # Backtest Search


    """
    def __init__(self):
        self._processor: Optional[Processor] = None
        self._create = CreateAccessor()
        self._update = UpdateAccessor()
        self._remove = RemoveAccessor()
        self._info = InfoAccessor()

    
    # @property
    # def processor(self):
    #     if self._processor is None:
    #         raise AttributeError("Injected processor hasn't been set. ")
    #     return self._processor

    # @processor.setter
    # def processor(self, _processor):
    #     self._processor = _processor

    @property
    def create(self):
        self._create.processor = self.processor
        return self._create
    
    @property
    def update(self):
        self._update.processor = self.processor
        return self._update
    
    @property
    def remove(self):
        self._remove.processor = self.processor
        return self._remove

    @property
    def info(self):
        self._info.processor = self.processor
        return self._info


def main_testing():
    pass

if __name__ == "__main__":
    main_testing()


import maya
import json
import time
import uuid
import datetime
from typing import Union, Optional, Any
from dataclasses import dataclass, field, fields
from jamboree.utils.core import consistent_hash
from addict import Dict as ADict


class TimeRange:
    """
        TimeRange
        ---
        A parameter to be used in remote backtesting. 
    """
    def __init__(self,
                 start: Union[str, float, int, datetime.datetime],
                 end: Union[str, float, int] = time.time(),
                 timezone: str = "UTC"):
        self.timezone = timezone
        self.start = self.HandleTimeParam(start)
        self.end = self.HandleTimeParam(end)
        self.interval = maya.MayaInterval(start=self.start, end=self.end)

    def HandleTimeParam(self, param: Union[str, float, int, datetime.datetime]):
        """Convert the parameter into a maya object. Generic time manipulation.

        Arguments:
            param {Union[str, float, int]} -- The time object epoch or string version of time

        Raises:
            ValueError: [description]

        Returns:
            [type] -- [description]
        """
        if isinstance(param, float) or isinstance(param, int):
            current_time = maya.MayaDT(param).datetime(
                to_timezone=self.timezone)
            prior_time = maya.MayaDT.from_datetime(current_time)
            return prior_time
        elif isinstance(param, str):
            return maya.when(param, timezone=self.timezone)
        elif isinstance(param, datetime.datetime):
            return maya.MayaDT.from_datetime(param)
        elif isinstance(param, datetime.date):
            return maya.when(param.isoformat(), timezone=self.timezone)
        print(type(param))
        print(isinstance(param, ))
        raise ValueError(
            f"Param is not the an acceptable datatype: [str, float, int] {type(param)}")

    def to_epoch_addict(self):
        epoch_dict = ADict()
        epoch_dict.start = self.start.epoch
        epoch_dict.end = self.end.epoch
        epoch_dict.duration = self.interval.duration
        return epoch_dict

    def to_epoch(self):
        """To Epoch
        
        Create an epoch addict dictionary carried with all of the necessary epoch information to create a backtest
        """
        epoch_dict = ADict()
        epoch_dict.start = self.start.epoch
        epoch_dict.end = self.end.epoch
        epoch_dict.duration = self.interval.duration
        return epoch_dict


class MovementSpecification:
    """ Use to deal with step information and lookback information """
    def __init__(self,
                 microseconds: float = 0,
                 seconds: float = 0,
                 minutes: float = 0,
                 hours: float = 0,
                 days: float = 0,
                 weeks: float = 0):
        self.microseconds = microseconds
        self.seconds = seconds
        self.minutes = minutes
        self.hours = hours
        self.days = days
        self.weeks = weeks

    def verify(self) -> bool:
        """Verifies that not all specified fields are not zero

        Returns:
            bool -- True if none are negative AND at least one is greater than 0
        """

        fields = [
            self.microseconds, self.seconds, self.minutes, self.hours,
            self.days, self.weeks
        ]
        return all(x >= 0 for x in fields) and any(x > 0 for x in fields)
    

    def to_dict(self):
        assert self.verify(), "At least one of the fields needs to be greater than 0, and all of them can't be negative"
        return {
            "microseconds": self.microseconds,
            "seconds": self.seconds,
            "minutes": self.minutes,
            "hours": self.hours,
            "days": self.days,
            "weeks": self.weeks
        }

    def to_json(self) -> str:
        """to_json
        
        Use to send complex information to other parts of the system. 

        Returns:
            str -- return a json string
        """
        return json.dumps(self.to_dict())

    def to_epoch(self):
        current_time = maya.now()
        next_time = maya.now()
        next_time = next_time.add(**self.to_dict())
        return next_time._epoch - current_time._epoch


import uuid
import maya
import random
from cytoolz.curried import pipe, map, filter, get
from cytoolz.itertoolz import unique

from typing import Optional, List
from jamboree import Jamboree
from jamboree.base.processors.abstracts import Processor
from jamboree.handlers.abstracted.search import ParameterizedSearch
from jamboree.handlers.default.search import BaseSearchHandler
from see137.search.prototype import SearchFacade


from loguru import logger



class UniverseAssetRelationshipManagement(BaseSearchHandler):
    """ 
        Controls the relationship between assets and universes. Use to find all universes by asset, and all assets by universe.
    """
    def __init__(self):
        super().__init__()
        self.entity = "universe_asset"
        self.dreq = {"universe": str, "asset": str}

class UniverseAssetCreate(SearchFacade):
    # """ 
    #     Used to find stategies and universes by asset or universe. 

    #     - Get the 

    #     ```py
    #         # Match a strategy with a universe and asset
    #         backtest_manager.create.universe_asset.init(universe: str, asset:str)
    #     ```
    # """
    def __init__(self):
        super().__init__(UniverseAssetRelationshipManagement)
    
    def init(self, universe: str, asset:str):
        universe_mix_id = self.Create(no_overwrite_reqs=True, universe=universe, asset=asset)
        return universe_mix_id

    
    
class UniverseAssetUpdate(SearchFacade):
    pass
    # """ 
    #     Use to find backtest information by episodes, and episodes by backtest

    #     ```py
    #     ```
    # """
    def __init__(self):
        super().__init__(UniverseAssetRelationshipManagement)
        

    


class UniverseAssetInfo(SearchFacade):
    """ 
        Use to find backtest information by episodes, and episodes by backtest

        - Is there any backtest that's running right now?
        - Find all of the unique universes

        ```py
            # Get the difference
            backtest_manager.info.episodes.by_backtest(backtest_id)
            
        ```
    """
    def __init__(self):
        super().__init__(UniverseAssetRelationshipManagement)
    
    @property
    def universes(self) -> List[dict]:
        """ get all episodes """
        universes = self.Find("*")
        unis = pipe(universes, map(lambda uni: uni['universe']), unique, list)
        return unis
    
    @property
    def assets(self) -> List[dict]:
        """ Get all active episodes"""
        assets = self.Find("*")
        unis = pipe(assets, map(lambda uni: uni['asset']), unique, list)
        return unis



    @property
    def total_universes(self) -> int:
        return len(self.universes)


    
    @property
    def total_assets(self) -> int:
        return len(self.assets)

    
    def universes_by_asset(self, assetid:str):
        """ Get all strategies by asset ID """
        strategy_list = self.Find(asset=assetid)
        strats = pipe(strategy_list, map(lambda uni: uni['universe']), unique, list)
        return strats
        

    
    def asset_by_universe(self, universe:str):
        assets = self.Find(universe=universe)
        _assets = pipe(assets, map(lambda assss: assss['asset']), unique, list)
        return _assets
    
    # def assets_by_user(self, user:str):
    #     """ Get all of the observed assets by user """
    #     assets = self.Find(user=user)
    #     _assets = pipe(assets, map(lambda assss: assss['asset']), unique, list)
    #     return _assets

    # def universes_by_user(self, user:str):
    #     """ Get all of the created universes by user. """
    #     strategy_list = self.Find(user=user)
    #     strats = pipe(strategy_list, map(lambda uni: uni['universe']), unique, list)
    #     return strats
    
    # def by_user(self, user:str):
    #     """ Get all records by user."""
    #     strategy_list = self.Find(user=user)
    #     return strategy_list

class UniverseAssetsRemove(SearchFacade):
    """ 
        
    """
    def __init__(self):
        super().__init__(UniverseAssetRelationshipManagement)
    

    def by_universe(self, universe:str):
        """ Remove episode by id"""
        self.Remove(universe=universe)

    def by_asset(self, asset:str):
        self.Remove(asset=asset)


def manage_assets():
    universe_id = uuid.uuid4().hex
    asset_ids = [uuid.uuid4().hex for x in range(200)]
    
    proc = Jamboree()
    universe_create =    UniverseAssetCreate()
    universe_update =    UniverseAssetUpdate()
    universe_info =      UniverseAssetInfo()
    universe_remove =    UniverseAssetsRemove()
    
    universe_create.processor = proc

    universe_update.processor = proc
    universe_info.processor = proc
    universe_remove.processor = proc
    universe_remove.RemoveAll()
    for assetid in asset_ids:
        universe_create.init(universe_id, assetid)
        logger.warning(universe_info.assets)
        logger.success(universe_info.universes)

if __name__ == "__main__":
    manage_assets()


from see137.parameters import Parameter
from jamboree.utils.core import consistent_hash

class StopLossParameter(Parameter):
    def __init__(self, complex_id:str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.comp_id = complex_id
        self.stop_loss_params = []
        # self.
    
    def extract_params(self, **kwargs):
        if len(self.stop_loss_params) == 0: return
        if len(kwargs) == 0: return
        for k, v in kwargs.items():
            if k in self.stop_loss_params:
                self.__dict__[k] = v
    
    
    def verify(self):
        pass

        


import time
import random
import uuid
import copy
from typing import List, Dict, Any, Set
from loguru import logger
from jamboree import Jamboree
from jamboree.handlers.default import DBHandler
from jamboree.handlers.default import MultiDataManagement
from jamboree.utils.helper import Helpers
from pprint import pprint


helpers = Helpers()


class RequirementsHandler(DBHandler):
    """ 
        # RequirementsHandler
        ---

        This manages the required items before moving onto the next step. 

        This is poorly written. Will likely rewrite it for more accuracy and speed.
    """

    def __init__(self):
        super().__init__()
        self.entity = "requirements"
        self.required = {
            "name": str,
            "live": bool,
            "episode": str
        }
        self['episode'] = "normal"
        self['live'] = True
        self.is_unique = False

        self._data = MultiDataManagement()
        self._episode: str = uuid.uuid4().hex
        self._live: bool = False
        self._assets: List[Dict[Any, Any]] = []
        self._items_compressed: Set[str] = set()
        self._reporting = {}

    @property
    def episode(self):
        return self._episode

    @episode.setter
    def episode(self, _episode):
        self._episode = _episode

    @property
    def live(self):
        return self._live

    @live.setter
    def live(self, _live):
        self._live: bool = _live

    @property
    def sources(self):
        """ Using multi-data for it validity checks. """
        name = self['name']
        episode = self['episode']
        live = self['live']
        self._data['name'] = f"{name}-{episode}-{live}"
        self._data['category'] = "assistance"
        self._data['subcategories'] = {}
        self._data['submetatype'] = "requirement"
        self._data['abbreviation'] = "REQMAN"
        self._data.episode = episode
        self._data.live = live
        self._data.processor = self.processor
        return self._data

    def unique(self):
        """ Get the requirements specified to the episode and live status"""
        unique_item = RequirementsHandler()
        unique_item['name'] = self['name']
        unique_item['episode'] = self.episode
        unique_item['live'] = self.live
        unique_item.assets = self.assets
        unique_item.episode = self.episode
        unique_item.processor = self.processor
        unique_item.is_unique = True
        unique_item.reset()
        return unique_item

    def asset_update(self, items: List[Dict[Any, Any]], command="add"):
        if len(items) == 0:
            return

        if command not in ["add", "sub"]:
            return

        _current_items = set()

        [_current_items.add(helpers.generate_hash(x)) for x in items]
        old_items = self.items
        if command == "add":
            old_items = old_items.union(_current_items)
        else:
            old_items = old_items.difference(_current_items)

        self.items = old_items
        # self.assets = self.decomp_items
        self.update()


    @property
    def assets(self):
        return self._assets

    @assets.setter
    def assets(self, _assets):
        self._assets = _assets
        self.assets_to_items()

    @property
    def items(self) -> Set:
        return self._items_compressed

    @items.setter
    def items(self, _items: Set[str]):
        self._items_compressed = _items

    @property
    def decomp_items(self):
        if len(self.items) == 0:
            return []
        return [helpers.hash_to_dict(x) for x in self.items]

    @property
    def is_valid(self):
        self.items = set(self._reporting.keys())
        self._general_update()

        report_values = list(self._reporting.values())
        if len(report_values) == 0:
            return False
        return all(report_values)


    def assets_to_items(self):
        _item_list = set()
        for asset in self.assets:
            _item_list.add(helpers.generate_hash(asset))
        self._items_compressed = _item_list

    def _unique_update(self):
        """ Adds the episode if this is an episode specific requirements """

        asset_list = []
        for asset in self.assets:
            if 'subcategories' in asset:
                asset['subcategories']['episode'] = self.episode
            else:
                asset['episode'] = self.episode
            asset_list.append(asset)
        self.assets = asset_list

    def default_reporting(self):
        self.prune_reports()

        for item in self.items:
            self._reporting[item] = self._reporting.get(item, False)

    def prune_reports(self):
        current_reports = copy.copy(self._reporting)
        for key in current_reports:
            if key not in self.items:
                self._reporting.pop(key, None)

    def reset_reports(self):
        """ Set the reporting to False and save"""
        for key in self._reporting.keys():
            self._reporting[key] = False
        self.update()

    def _general_update(self):

        self.assets = self.decomp_items
        self.default_reporting()

    """ Core functions """
    def report(self, asset: Dict[Any, Any]):
        """ Report that we've passed over an asset """
        self.load()
        itemized = helpers.generate_hash(asset)
        self._reporting[itemized] = True
        self.update()

    def load(self):
        """ Load all of the assets and the status of those assets as well """
        asset_set = self.last()

        report_set = self.last(alt={"detail": "report"})
        self.assets = asset_set.get("assets", [])
        self._reporting = report_set.get("report", {})
        if self.is_unique:
            self._unique_update()
            self.assets_to_items()
        self._general_update()

    def update(self):
        if self.is_unique:
            self._unique_update()
            self.assets_to_items()
        self._general_update()

        data = {"assets": self.assets,
                "time": time.time(), "timestamp": time.time()}

        report = {"report": self._reporting,
                  "time": time.time(), "timestamp": time.time()}
        self.save(data)
        self.save(report, alt={"detail": "report"})

    def count_assets(self) -> int:
        """Get count of assets that are required in this session.

        Returns:
            int -- Required Asset Cardinality
        """
        # This should be cached at some point
        current_count = self.get_single(alt={"detail": "cardinal_count"})
        if bool(current_count):
            return
        # logger.success(current_count)
        return len(self.assets)

    def count_cardinality(self):
        return self.count(alt={"detail": "cardinal"})

    def report_cardinality(self, asset: dict):
        self.save({"asset": asset}, alt={"detail": "cardinal"})

    def flush_cardinality(self):
        self.delete_all(alt={"detail": "cardinal"})

    @property
    def is_valid_cardinality(self) -> bool:
        """Detects if the asset cardinality matches the current number of assets.

        Returns:
            bool -- [description]
        """
        current_count = self.count_cardinality()
        total_asset_count = self.count_assets()
        # logger.info(current_count)
        # logger.info(total_asset_count)
        return (current_count >= total_asset_count)

    def reset(self):
        self.sources.reset()
        req_count = self.count()
        if req_count == 0:
            pass
            # logger.debug("Updating for the first time")

            self.update()
        else:
            self.load()

    def info(self):
        return {
            "episode": self.episode,
            "is_unique": self.is_unique
        }


if __name__ == "__main__":

    all_assets = ["BTC", "XTZ", "ETH", "ATX",
                  "XRP", "BCH", "BSV", "LTC", "EOS"]

    jambo = Jamboree()
    reqhandler = RequirementsHandler()
    reqhandler.processor = jambo
    # The name is on rick and morty so it's not perverted. I can prove it
    reqhandler['name'] = "PoopybuttholeS"
    reqhandler.assets = [
        {
            "name": x,
            "category": "market",
            "subcategories": {"exchange": "fake_exchange"}
        }
        for x in all_assets
    ]
    reqhandler.reset()
    new_items = [
        {
            "name": x,
            "category": "market",
            "subcategories": {"exchange": "fake_exchange"}
        }
        for x in ["ONE", "TWO", "THREE", "FOUR", "BTC", "ATX"]
    ]
    reqhandler.asset_update(new_items, command="sub")
    reqhandler.reset()

    assets = reqhandler.assets
    while True:
        current_asset = random.choice(assets)
        # logger.info(current_asset)
        reqhandler.report_cardinality(current_asset)
        # logger.success(reqhandler.is_valid_cardinality)
        if reqhandler.is_valid_cardinality:
            logger.debug("Episode step is done")
            logger.success("Pushing the time step forward")
            reqhandler.flush_cardinality()
        else:
            logger.warning("Not done yet")
            logger.error("Handling next step")



import uuid
import maya

from typing import Optional
from jamboree import Jamboree
from jamboree.base.processors.abstracts import Processor
from see137.search.backtesting.topics import backtests as bts
from see137.search.backtesting.topics import episodes as eps
from see137.search.backtesting.topics import strategy as sst
from see137.search.backtesting.topics import universe as uni
from see137.search.backtesting.topics.asset import universal as unias

class CreateAccessor:
    def __init__(self):
        self._processor: Optional[Processor] = None
        self._backtest = bts.BacktestCreate()
        self._episodes = eps.EpisodeCreate()
        self._strategy = sst.StrategyCreate()
        self._universe = uni.UniverseCreate()
        self._unias = unias.UniverseAssetCreate()
        # self._universe_detail = univo.UniverseDetailCreate()

    @property
    def processor(self):
        if self._processor is None:
            raise AttributeError("Processor hasn't been set for create calls in search.")
        return self._processor

    @processor.setter
    def processor(self, _processor):
        self._processor = _processor

    @property
    def backtest(self):
        self._backtest.processor = self.processor
        return self._backtest
    
    @property
    def episodes(self):
        self._episodes.processor = self.processor
        return self._episodes
    
    @property
    def strategy(self):
        self._strategy.processor = self.processor
        return self._strategy

    @property
    def universe(self):
        self._universe.processor = self.processor
        return self._universe
    
    @property
    def universe_asset(self):
        self._unias.processor = self.processor
        return self._unias


class InfoAccessor:
    def __init__(self):
        self._processor: Optional[Processor] = None
        self._backtest = bts.BacktestInfo()
        self._episodes = eps.EpisodeInfo()
        self._strategy = sst.StrategyInfo()
        self._universe = uni.UniverseInfo()
        self._unias = unias.UniverseAssetInfo()

    @property
    def processor(self):
        if self._processor is None:
            raise AttributeError("Processor hasn't been set for create calls in search.")
        return self._processor

    @processor.setter
    def processor(self, _processor):
        self._processor = _processor

    @property
    def backtest(self):
        self._backtest.processor = self.processor
        return self._backtest
    
    @property
    def episodes(self):
        self._episodes.processor = self.processor
        return self._episodes
    
    @property
    def strategy(self):
        self._strategy.processor = self.processor
        return self._strategy

    @property
    def universe(self):
        self._universe.processor = self.processor
        return self._universe
    
    @property
    def universe_asset(self):
        self._unias.processor = self.processor
        return self._unias


class RemoveAccessor:
    def __init__(self):
        self._processor: Optional[Processor] = None
        self._backtest = bts.BacktestRemove()
        self._episodes = eps.EpisodeRemove()
        self._strategy = sst.StrategyRemove()
        self._universe = uni.UniverseRemove()
        self._unias = unias.UniverseAssetsRemove()


    @property
    def processor(self):
        if self._processor is None:
            raise AttributeError("Processor hasn't been set for create calls in search.")
        return self._processor

    @processor.setter
    def processor(self, _processor):
        self._processor = _processor


    @property
    def backtest(self):
        self._backtest.processor = self.processor
        return self._backtest
    
    @property
    def episodes(self):
        self._episodes.processor = self.processor
        return self._episodes
    
    @property
    def strategy(self):
        self._strategy.processor = self.processor
        return self._strategy

    @property
    def universe(self):
        self._universe.processor = self.processor
        return self._universe
    
    @property
    def universe_asset(self):
        self._unias.processor = self.processor
        return self._unias


class UpdateAccessor:
    def __init__(self):
        self._processor: Optional[Processor] = None
        self._backtest = bts.BacktestUpdate()
        self._episodes = eps.EpisodeUpdate()
        self._strategy = sst.StrategyUpdate()
        self._universe = uni.UniverseUpdate()
        self._unias = unias.UniverseAssetUpdate()

    @property
    def processor(self):
        if self._processor is None:
            raise AttributeError("Processor hasn't been set for create calls in search.")
        return self._processor

    @processor.setter
    def processor(self, _processor):
        self._processor = _processor
    

    @property
    def backtest(self):
        self._backtest.processor = self.processor
        return self._backtest
    
    @property
    def episodes(self):
        self._episodes.processor = self.processor
        return self._episodes
    
    @property
    def strategy(self):
        self._strategy.processor = self.processor
        return self._strategy

    @property
    def universe(self):
        self._universe.processor = self.processor
        return self._universe
    
    @property
    def universe_asset(self):
        self._unias.processor = self.processor
        return self._unias


import uuid
import maya
import random
from cytoolz.curried import pipe, map, filter, get
from cytoolz.itertoolz import unique

from typing import Optional, List
from jamboree import Jamboree
from jamboree.base.processors.abstracts import Processor
from jamboree.handlers.abstracted.search import ParameterizedSearch
from jamboree.handlers.default.search import BaseSearchHandler
from see137.search.prototype import SearchFacade

from loguru import logger

class ProcedureSearch(BaseSearchHandler):
    def __init__(self):
        super().__init__()
        self.entity = "asset_complex"
        self.dreq = {"typing":str, "asset": str, "params": dict, "next_delta": float, "interval": float}

class UniverseAssetCreate(SearchFacade):
    """ 
        Used to find stategies and universes by asset or universe. 

        - Get the 

        ```py
            # Match a strategy with a universe and asset
            backtest_manager.create.universe_asset.init(universe: str, asset:str)
        ```
    """
    def __init__(self):
        super().__init__(ProcedureSearch)
    
    def init(self, universe: str, asset:str):
        universe_mix_id = self.Create(no_overwrite_reqs=True, universe=universe, asset=asset)
        return universe_mix_id

    
    
class UniverseAssetUpdate(SearchFacade):
    def __init__(self):
        super().__init__(ProcedureSearch)
        

    


class UniverseAssetInfo(SearchFacade):
    def __init__(self):
        super().__init__(ProcedureSearch)
    @property
    def universes(self) -> List[dict]:
        """ get all episodes """
        universes = self.Find("*")
        unis = pipe(universes, map(lambda uni: uni['universe']), unique, list)
        return unis
    
    @property
    def assets(self) -> List[dict]:
        """ Get all active episodes"""
        assets = self.Find("*")
        unis = pipe(assets, map(lambda uni: uni['asset']), unique, list)
        return unis



    @property
    def total_universes(self) -> int:
        return len(self.universes)


    
    @property
    def total_assets(self) -> int:
        return len(self.assets)

    
    def universes_by_asset(self, assetid:str):
        """ Get all strategies by asset ID """
        strategy_list = self.Find(asset=assetid)
        strats = pipe(strategy_list, map(lambda uni: uni['universe']), unique, list)
        return strats
        

    
    def asset_by_universe(self, universe:str):
        assets = self.Find(universe=universe)
        _assets = pipe(assets, map(lambda assss: assss['asset']), unique, list)
        return _assets

class UniverseAssetsRemove(SearchFacade):
    """ 
        
    """
    def __init__(self):
        super().__init__(ProcedureSearch)
    

    def by_universe(self, universe:str):
        """ Remove episode by id"""
        self.Remove(universe=universe)

    def by_asset(self, asset:str):
        self.Remove(asset=asset)


def manage_assets():
    pass
    # universe_id = uuid.uuid4().hex
    # asset_ids = [uuid.uuid4().hex for x in range(200)]
    
    # proc = Jamboree()
    # universe_create =    UniverseAssetCreate()
    # universe_update =    UniverseAssetUpdate()
    # universe_info =      UniverseAssetInfo()
    # universe_remove =    UniverseAssetsRemove()
    
    # universe_create.processor = proc

    # universe_update.processor = proc
    # universe_info.processor = proc
    # universe_remove.processor = proc
    # universe_remove.All()
    # for assetid in asset_ids:
    #     universe_create.init(universe_id, assetid)
    #     logger.warning(universe_info.assets)
    #     logger.success(universe_info.universes)

if __name__ == "__main__":
    manage_assets()


import uuid
from loguru import logger
from typing import List, Dict, Tuple, Any
from addict import Dict as ADict
from see137.common import ComponentAbstract, AddictComponent
from see137.search.backtesting import BacktestManager


class AddictWithManagerComponent(AddictComponent):
    """
        A component that centers around using dictionaries and addict.
    """
    def __init__(self):
        super().__init__()
        self.__manager = BacktestManager()
    
    @property
    def manager(self):
        self.__manager.processor = self.processor
        return self.__manager

if __name__ == "__main__":
    comp = AddictWithManagerComponent()
    sample_param = ADict(episode=uuid.uuid4().hex)
    print(comp.step(sample_param, asset=uuid.uuid4().hex))


import numpy as np
import pandas as pd
from empyrical import roll_sortino_ratio, roll_max_drawdown, roll_alpha_beta, roll_sharpe_ratio

def create_sharpe_ratio(returns,periods=20):
    return (
        np.sqrt(periods) *
        np.mean(returns) /
        np.std(returns)
    )


def create_drawdowns(equity_curve):
    high_watermark = [0]
    curve_index = equity_curve.index
    drawdown = pd.Series(index=curve_index)
    duration = pd.Series(index=curve_index)

    for i in range(1, len(curve_index)):
        current_high_watermark = max(
            high_watermark[i-1],
            equity_curve[i]
        )
        high_watermark.append(current_high_watermark)

        drawdown[i] = high_watermark[i] - equity_curve[i]
        duration[i] = 0 if drawdown[i] == 0 else duration[i-1] + 1

    return drawdown.max(), duration.max()


def get_rolling_performance_metrics(series:list, _window=5):
    if len(series) < _window:
        return {
            "sortino": [],
            "sharpe": [],
            "alpha_beta": [],
            "drawdown": []
        }
    

    sortino = roll_sortino_ratio(series, window=_window)
    sharpe = roll_sharpe_ratio(series, window=_window)
    alpha_beta = roll_alpha_beta(series, window=_window)
    drawdown = roll_max_drawdown(series, window=_window)
    
    return {
        "sortino": sortino,
        "sharpe": sharpe,
        "alpha_beta": alpha_beta,
        "drawdown": drawdown
    }


"""Backtest Environment

Use environment to control backtests remotely inside of the Kafka Queues.
"""


import uuid

import maya
from addict import Dict as ADict
from faker import Faker
from jamboree import Jamboree
from loguru import logger
import json
from see137.parameters import (Parameter, PortfolioParameter,
                               StrategyParameter, TimeParameter,
                               UniverseParameter, UserParameter)
from see137.parameters.fields import (CurrencyFields, ExchangeFields,
                                      MovementSpecification, TimeRange)
from see137.pipelines.setup_backtest import CreateBacktestPipeline
from see137.search.backtesting import BacktestManager
import requests
logger.disable("jamboree")
# logger.disable("see137")


class BacktestEnvironment:
    def __init__(self,
                 name: str,
                 description: str,
                 time_param: TimeParameter = None,
                 portfolio_param=None,
                 universe_param=None,
                 user_param=None,
                 strategy_param=None):
        self.backtest_identity = None
        self.bid = None

        self.name = name
        self.description = description
        # Search for all backtest

        # Set multi data config
        self.time_param = time_param
        self.portfolio_param = portfolio_param
        self.universe_param = universe_param
        self.user_param = user_param
        self.strategy_param = strategy_param
        self.__pipeline = CreateBacktestPipeline()
        self.verify()

        # Private To Be Called by Property
        self.__backtest_manager = BacktestManager()

    @property
    def processor(self):
        """Processor

        Returns
        -------
        Processor
            The jamboree object that processes everything.

        Raises
        ------
        AttributeError
            Raises when the Jamboree flywheel object doesn't work.
        """
        if self._processor is None:
            raise AttributeError("Injected processor hasn't been set. ")
        return self._processor

    @processor.setter
    def processor(self, _processor):
        self._processor = _processor

    @property
    def manager(self):
        self.__backtest_manager.processor = self.processor
        return self.__backtest_manager

    @property
    def pipeline(self):
        """Backtest Pipeline

        Gets the backtest creation pipeline.
        Verification and creation of a backtest to be picked up in the system.

        Returns
        -------
        AddictPipeline
            CreateBacktestPipeline - Backtest creation pipeline.
        """
        self.__pipeline.processor = self.processor
        return self.__pipeline

    def monitor(self):
        """ What we use to aggrgate the information about the backtest."""
        # Once we have an id generated and appended
        return None

    def performance(self):
        """ Get the performance of the backtest """
        return None

    def status(self):
        """ Get the current status of the backtest. Is it finished? How much time left?"""
        return None

    def resources(self):
        return None

    def extract(self) -> dict:
        extract_addict = ADict()
        extract_addict.name = self.name
        extract_addict.description = self.description
        extract_addict.user = self.user_param.user

        extract_addict.base = self.portfolio_param.base
        extract_addict.initial = self.portfolio_param.initial
        extract_addict.exchange = self.portfolio_param.exchange

        extract_addict.start = float(self.time_param.start)
        extract_addict.end = float(self.time_param.end)
        extract_addict.step = self.time_param.step
        extract_addict.lookback = self.time_param.lookback

        extract_addict.strategy = self.strategy_param.strategy
        extract_addict.universe = self.universe_param.universe

        return extract_addict.to_dict()

    def verify(self):
        """Verify
        Verifies that all of the parameters are available.

        Raises
        ------
        AttributeError
            One of the attributes is missing.
        TypeError
            One of the attributes isn't a parameter system.
        """
        all_conf = [("Time Parameter", self.time_param),
                    ("Portfolio Parameter", self.portfolio_param),
                    ("Universe Parameter", self.universe_param),
                    ("User Parameter", self.user_param),
                    ("Strategy Parameter", self.strategy_param)]
        for index in range(len(all_conf)):
            key, value = all_conf[index]
            if value is None:
                raise AttributeError(
                    f"{key} is missing. Make sure it is not None")
            if not isinstance(value, Parameter):
                raise TypeError(f"{key} needs to be a Parameter class type")

    @staticmethod
    def by_id(_id: str):
        """Get backtest information by id.

        Pulls the information from the database if it exist.

        Example:

        ::
            interval = TimeRange(start=start, end=end)
            steps = MovementSpecification(days=3)
            lookback = MovementSpecification(weeks=4)

            time_parameter = TimeParameter(interval=interval,
                                        steps=steps,
                                        lookback=lookback)

            currency = CurrencyFields()
            exchange = ExchangeFields("coinbase")

            portfolio_parameter = PortfolioParameter(currency=currency,
                                                    exchange=exchange)
            user_parameter = UserParameter("userid")
            universe_parameter = UniverseParameter("universeid")
            strategy_parameter = StrategyParameter("strategyid")


        Arguments:
            _id {str} -- The id of the backtest we're trying to retrieve. 

        Returns:
            BacktestEnvironment -- return a valid instance of the backtest environment we're looking for.
        """
        return None
        # return BacktestEnvironment

    def create(self):
        extract = self.extract()
        backtest_info = ADict()
        backtest_info.config = extract
        _info = self.pipeline.step(backtest_info)
        is_success = (not _info.failed)
        message = str(_info.msg)
        backtestid = str(_info.backtestid)

        if not is_success:
            raise Exception(message)
        
        # Set the backtestid to be used elsewhere
        self.bid = backtestid

    def connect(self):
        """ Input settings of a place to send the start command to remote service"""
        try:
            response = requests.get("http://localhost:6066/episodes/connect")
            logger.warning(response.status_code)
            response.json()
            return True
        except json.decoder.JSONDecodeError:
            logger.error("FAILED CONNECTION")
            return False
        except Exception as exe:
            # Using a catch all for some otehr error that might pop up
            logger.exception(exe)
            return False
        

    def push(self):
        try:
            response = requests.post("http://localhost:6066/episodes/init", json={"backtest": self.bid})
            jresponse = response.json()
            is_successful = jresponse.get("success", False)
            if not is_successful:
                return False
            logger.success(f"Backtest = {self.bid} was successfully started")
            return True
        except json.decoder.JSONDecodeError:
            return False
        except Exception:
            return False

    @logger.catch(reraise=True)
    def start(self):
        if self.connect():
            self.create()
            self.push()


def main():
    # We'd get this information from the user in real life
    fake = Faker()
    name = fake.company()
    description = fake.sentence()
    start = fake.date_between(start_date="-10y")
    end = maya.now().epoch

    interval = TimeRange(start=start, end=end)
    steps = MovementSpecification(days=3)
    lookback = MovementSpecification(weeks=4)

    time_parameter = TimeParameter(interval=interval,
                                   steps=steps,
                                   lookback=lookback)

    currency = CurrencyFields()
    exchange = ExchangeFields("coinbase")
    portfolio_parameter = PortfolioParameter(currency=currency,
                                             exchange=exchange)
    user_parameter = UserParameter(uuid.uuid4().hex)
    universe_parameter = UniverseParameter("92458de9072d46c8be1883745bc24b37")
    strategy_parameter = StrategyParameter("53970e2f7fae4990978b61c8cd0e5b57")
    injectable_processor = Jamboree()
    backtest_env = BacktestEnvironment(name,
                                       description,
                                       time_param=time_parameter,
                                       portfolio_param=portfolio_parameter,
                                       universe_param=universe_parameter,
                                       user_param=user_parameter,
                                       strategy_param=strategy_parameter)
    backtest_env.processor = injectable_processor
    backtest_env.start()


if __name__ == "__main__":
    main()



import uuid
import maya

from typing import Optional, List
from jamboree import Jamboree
from jamboree.base.processors.abstracts import Processor
from jamboree.handlers.abstracted.search import ParameterizedSearch
from jamboree.handlers.default.search import BaseSearchHandler
from see137.search.prototype import SearchFacade
from loguru import logger



class _CoreBacktestSearch(BaseSearchHandler):
    """ 
        # Core Backtest Search  

        Use to find backtesting information.

        Examples:

        ```py

            backtest_manager.create.backtest.init(**kwargs)


            backtest_manager.update.backtest.general(id, **kwargs)
            backtest_manager.update.backtest.time(id, **kwargs)
            backtest_manager.update.backtest.portfolio(id, **kwargs)
            backtest_manager.update.backtest.description(id, **kwargs)
            backtest_manager.update.backtest.strategy(id, **kwargs)


            backtest_manager.info.backtest.by_id(id)
            backtest_manager.info.backtest.by_user(userid)
            backtest_manager.info.backtest.by_exchange(exchange)
            backtest_manager.info.backtest.by_strategy(strategy_id)
            backtest_manager.info.backtest.by_universe(universe_id)

            backtest_manager.info.backtest.start_before('...')
            backtest_manager.info.backtest.start_after('...')
            backtest_manager.info.backtest.start_between(start='...', end='...')

            backtest_manager.info.backtest.end_before('...')
            backtest_manager.info.backtest.end_after('...')
            backtest_manager.info.backtest.end_between(start='...', end='...')


            backtest_manager.remove.backtest.by_user(userid)
            backtest_manager.remove.backtest.by_exchange(exchange)
            backtest_manager.remove.backtest.by_strategy(strategy_id)
            backtest_manager.remove.backtest.by_universe(universe_id)

            backtest_manager.info.backtest.over_cutoff(backtestid, current_time)
        ```
    """

    def __init__(self):
        super().__init__()
        self.entity = "backtest_configuration"
        self.dreq = {
            "user": str,
            "name": str,
            "description": str,

            # Time series information.
            "start": float,
            "end": float,
            "step": str,
            "lookback": str,

            # Outside information
            "base": str,
            "initial": float,

            "exchange": str,

            # Outside information
            "strategy": str,
            "universe": str,
        }

class CoreBacktestSearch(SearchFacade):
    def __init__(self):
        super().__init__(_CoreBacktestSearch())
        self.must_have = ['name', 'description', 'user']
        


class BacktestCreate(CoreBacktestSearch):
    def __init__(self):
        super().__init__()
    
    # @logger.catch
    def init(self, **kwargs):
        backtest_id = self.Create(no_overwrite_reqs=True, **kwargs)
        return backtest_id

    

class BacktestUpdate(CoreBacktestSearch):
    def __init__(self):
        super().__init__()
        self.any_time: List[str] = ["start", "end", "step"]
        self.any_portfolio: List[str] = ["initial", "exchange", "universe"]
        self.any_description: List[str] = ["description"]
        self.any_strategy: List[str] = ["strategy"]
    
    def generalized(self, _id:str, **kwargs):
        self.UpdateID(_id, **kwargs)

    def time(self, _id:str, **kwargs):
        """ Update time information to a backtest. """
        if len(kwargs) == 0:
            return
        if any(anne in kwargs for anne in self.any_time):
            self.UpdateID(_id, **kwargs) #start_epoch

    def portfolio(self, _id:str, **kwargs):
        """ Add portfolio information for a backtest """
        if len(kwargs) == 0:
            return
        if any(anne in kwargs for anne in self.any_portfolio):
            self.UpdateID(_id, **kwargs) #start_epoch


    def strategy(self, _id:str, **kwargs):
        """ Set a strategy """
        if any(anne in kwargs for anne in self.any_strategy):
            self.UpdateID(_id, **kwargs)


class BacktestInfo(CoreBacktestSearch):
    """ 

        backtest_manager.info.backtest.by_id(id)
        backtest_manager.info.backtest.by_user(userid)
        backtest_manager.info.backtest.by_exchange(exchange)
        backtest_manager.info.backtest.by_strategy(strategy_id)
        backtest_manager.info.backtest.by_universe(universe_id)

        backtest_manager.info.backtest.start_before('...')
        backtest_manager.info.backtest.start_after('...')
        backtest_manager.info.backtest.start_between(start='...', end='...')

        backtest_manager.info.backtest.end_before('...')
        backtest_manager.info.backtest.end_after('...')
        backtest_manager.info.backtest.end_between(start='...', end='...')
    """
    def __init__(self):
        super().__init__()
    
    @property
    def count(self) -> int:
        return 0
    
    def generalized(self, term:str=None, **kwargs):
        return self.Find(general=term, **kwargs)

    def by_id(self, _id:str, **kwargs):
        return self.FindById(_id)

    def by_user(self, user:str):
        return self.Find(user=user)

    def by_exchange(self, exchange:str):
        return self.Find(exchange=exchange)

    def by_strategy(self, strategy:str):
        return self.Find(strategy=strategy)

    def by_universe(self, universe:str):
        return self.Find(universe=universe)

    def over_cutoff(self, backtest_id:str, current_time:float):
        backtest_info = self.FindById(backtest_id)
        if not bool(backtest_info):
            return True
        


class BacktestRemove(CoreBacktestSearch):
    def __init__(self):
        super().__init__()
    
    def generalized(self, term:str=None, **kwargs):
        if term is not None:
            self.general = term
        
        self.Remove(**kwargs)

    def by_id(self, _id:str):
        pass

    def by_user(self, user:str):
        return self.Remove(user=user)

    def by_exchange(self, exchange:str):
        return self.Remove(exchange=exchange)

    def by_strategy(self, strategy:str):
        return self.Remove(strategy=strategy)

    def by_universe(self, universe:str):
        return self.Remove(universe=universe)

def manage_episodes():
    from faker import Faker

    Faker.seed(444)


    fake = Faker()
    jampack = Jamboree()
    backtest_create =   BacktestCreate()
    backtest_update =   BacktestUpdate()
    backtest_info =     BacktestInfo()
    backtest_remove =   BacktestRemove()


    backtest_create.processor = jampack
    backtest_update.processor = jampack
    backtest_info.processor = jampack
    backtest_remove.processor = jampack


    while True:
        user_id = uuid.uuid4().hex
        name = f"{fake.first_name()} {fake.last_name()}"
        description = str(fake.sentence())
        backtest_id = backtest_create.init(name=name, description=description, userid=user_id)
        logger.error(description)
        logger.success(backtest_info.by_id(backtest_id).get("description", ""))
        logger.debug(backtest_info.generalized(name=name))


if __name__ == "__main__":
    manage_episodes()


import copy
import random
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

import dask
import maya
from jamboree import Jamboree
from jamboree.handlers.default.data import DataHandler
from loguru import logger

from see137 import PortfolioHandler, RequirementsHandler
from see137.data import PriceGenerator
from see137.utils.trades.trade import DynamicTrade, Trade, TradeType


class SimulationCreator(object):
    """ 
        # Simulation Creator
        ---
        Create a simulation that we'll be able to run through the faust pipeline.

        * Requirements
        * Stochastic Data
        * Portfolios
        * TimeHandler
    """
    def __init__(self,
                 start_balance=10000,
                 session_id: str = uuid.uuid4().hex,
                 ecount=10,
                 exchange="fake_exchange",
                 **kwargs) -> None:
        self.session: str = session_id
        self.episode_count = ecount
        self.exchange = exchange
        self.start_balance = start_balance
        self.assets = [
            "USD_BTC", "USD_ATL", "USD_TRX", "USD_ETH", "USD_BCH", "USD_XRP",
            "USD_EOS", "USD_XTZ", "USD_LINK", "USD_ADA"
        ]

        # Generated hexes
        self.episodes = [uuid.uuid4().hex for i in range(self.episode_count)]
        self.user_id = uuid.uuid4(
        ).hex  # We're generating a fake user id for the time being
        #
        self._current_episode: Optional[str] = None
        self.processor = Jamboree()

        # Kwargs
        self.min_price = kwargs.get("min_price", 200)
        self.max_price = kwargs.get("max_price", 5000)
        self.varied = kwargs.get("is_varied", True)
        self.length = kwargs.get("length", 5000)

        # Blanks
        self.generated = None
        self.is_requirements_reset = False

        self.submetatype = "temporary_price"
        self.abbreviation = "GEN"

    @property
    def current_episode(self) -> str:
        if self._current_episode is None:
            return uuid.uuid4().hex
        return self._current_episode

    @current_episode.setter
    def current_episode(self, _episode: str):
        self._current_episode = _episode

    @property
    def pricing(self):
        data_handler = DataHandler()
        data_handler.processor = self.processor
        data_handler["category"] = self.category
        data_handler["subcategories"] = self.subcategories
        data_handler['submetatype'] = self.submetatype
        data_handler['abbreviation'] = self.abbreviation
        return data_handler

    @property
    def category(self):
        return "markets"

    @property
    def subcategories(self):
        return {
            "market": "stock",
            "country": "US",
            "sector": "faaaaake",
            "episode": self.current_episode
        }

    @property
    def required(self):
        sub = self.subcategories
        sub.pop("episode")
        base = {
            "category": self.category,
            "subcategories": sub,
            "submetatype": self.submetatype,
            "abbreviation": self.abbreviation
        }
        ret_items = []
        for asset in self.assets:
            base = copy.copy(base)
            base['name'] = asset
            ret_items.append(base)
        return ret_items

    @property
    def active(self):
        """ Get all of the active requirements"""
        return self.requirements.assets

    @property
    def sampled_portfolio(self):
        """ Return a sample portfolio for the user """
        assest_num = len(self.assets)
        num_sample = random.randint(2, assest_num)
        asset_sample = random.sample(self.assets, num_sample)
        return asset_sample

    @property
    def generator(self):
        """ Price generator """
        price_generator = PriceGenerator()
        price_generator.episodes = self.episodes
        price_generator.assets = self.assets
        price_generator.starting_min = self.min_price
        price_generator.starting_max = self.max_price
        price_generator.is_varied = self.varied
        price_generator.length = self.length
        return price_generator

    @property
    def portfolio(self):
        """ """
        portfolio = PortfolioHandler(start_balance=self.start_balance)
        portfolio.processor     = self.processor
        portfolio['session']    = self.session
        portfolio['exchange']   = self.exchange
        portfolio['user_id']    = self.user_id
        portfolio['episode']    = self.current_episode
        portfolio['live'] = False
        portfolio.reset()
        portfolio.is_sim = True
        portfolio.is_inner_allocation = False
        return portfolio

    @property
    def current_generated(self) -> dict:
        """ Get all of the prices for current episode """
        if self.generated is None:
            self.generator.generate_all_episodes()
        current_prices = self.generated
        return current_prices.get(self.current_episode, {})

    @property
    def requirements(self):
        requirements = RequirementsHandler()
        requirements.processor = self.processor
        requirements['name'] = f"general-{self.session}"
        requirements.episode = self.current_episode
        requirements.live = False
        requirements.reset()
        # if self.is_requirements_reset == False:

        #     self.is_requirements_reset = True
        return requirements

    def generate(self):
        self.generated = self.generator.generate_all_episodes()

        #

    def save_generated(self):
        """ Saving generated data for use in real life """
        logger.info("Saving generated data")
        current_pricing = self.pricing
        current_generated = self.current_generated
        for asset in self.assets:
            current_pricing['name'] = asset
            data = current_generated.get(asset)
            current_pricing.reset()
            current_pricing.store_time_df(data, is_bar=True)

    def load_requirements(self):
        required_items = self.required
        self.requirements.asset_update(required_items)
        # self.requirements.update()

    def load_assets(self):
        """ Loads the user's assets into the portfolio """
        user_sample = self.sampled_portfolio
        current_portfolio = self.portfolio
        for asset in user_sample:
            extras = {
                "submetatype": self.submetatype,
                "abbreviation": self.abbreviation
            }
            current_portfolio.add_asset(asset,
                                        self.subcategories,
                                        extras=extras)
        # logger.info(self.portfolio.pricing.sources)

    def load_unique_requirements(self):
        self.requirements.live = False
        self.requirements.episode = self.current_episode
        self.requirements.unique()
        # logger.info(self.portfolio.pricing.sources)

    def load_allocation(self):
        current_portfolio = copy.copy(self.portfolio)
        current_portfolio.step_price(forced=True)

    def adjust_time(self):
        """
            portfolio_hand.time.head = maya.now().add(weeks=25)._epoch
            portfolio_hand.time.change_stepsize(microseconds=0, days=1, hours=0)
            portfolio_hand.time.change_lookback(microseconds=0, weeks=25, hours=0)
        """

        self.portfolio.time.head = maya.now().add(weeks=5)._epoch
        self.portfolio.time.change_stepsize(microseconds=0, days=1, hours=0)
        self.portfolio.time.change_lookback(microseconds=0, weeks=5, hours=0)
        self.portfolio.pricing.sync()

    def start_debug(self):
        self.load_requirements()

    def start(self):
        """ Create all of the episodes"""
        self.load_requirements()
        self.generate()
        logger.info("Successfully starting session")
        generation_tasks = []
        for episode in self.episodes:
            self.current_episode = episode
            self.save_generated()
            self.adjust_time()
            self.load_assets()
            self.load_allocation()
            self.load_unique_requirements()


def process_rebalance(current_portfolio: PortfolioHandler,
                      current_requirements: RequirementsHandler):
    logger.debug("Stepping forward for the user")
    current_portfolio.time.step()
    current_portfolio.is_inner_allocation = True
    current_portfolio.step_price()
    current_portfolio.is_inner_allocation = False
    logger.error("Finished all assets in episode")


if __name__ == "__main__":
    simulation_creator = SimulationCreator(length=2000, ecount=2)
    simulation_creator.start()
    # executor = ThreadPoolExecutor()

    # Get all of the active assets in the simulation

    while True:
        is_done = False
        for episode in simulation_creator.episodes:
            simulation_creator.current_episode = episode
            current_portfolio = simulation_creator.portfolio
            current_requirements = simulation_creator.requirements.unique()
            active_assets = current_requirements.assets
            current_active = copy.copy(active_assets)
            while True:
                if not current_portfolio.pricing.is_next:
                    is_done = True
                    break

                active = current_active.pop()
                symbol = active.get('name')
                asset_msg = f"Processing the asset: {symbol} for episode: {episode} we'd pull data and process it."
                # logger.info(magenta(asset_msg))
                current_item = simulation_creator.pricing
                current_item['name'] = symbol
                current_item.episode = episode
                current_item.live = False
                current_item.reset()
                # Get the callocation and fill information here
                user_id = current_portfolio.user_id
                # logger.warning(f"User ID is {user_id}")
                current_portfolio.add_asset(symbol)
                current_item.closest_head()
                current_allocation = current_portfolio.current_allocation(symbol=symbol)
                current_limit = current_portfolio.current_pct_limit(symbol=symbol)
                is_not_default = "{:.21f}".format(float(current_allocation)) != '0.000000000000000000001'
                
                # if is_not_default:
                logger.warning((symbol, current_allocation, current_limit, episode))
                # logger.debug(current_limit)
                if current_portfolio.is_user_asset(active):
                    # logger.success(asset_msg)
                    if random.uniform(0, 1) < 0.1:
                        # Makes trade for portfolio
                        trade_type = random.choice([TradeType.LIMIT_BUY, TradeType.LIMIT_SELL])
                        dynamic_trade = DynamicTrade(symbol, trade_type)
                        dynamic_trade.percentage = random.normalvariate(0.3, 0.06)
                        current_portfolio.step(dynamic_trade)
                    else:
                        current_portfolio.step()
                current_requirements.report_cardinality(active)
                if current_requirements.is_valid_cardinality:
                    current_requirements.flush_cardinality()
                    current_portfolio.time.step()

                    logger.error("Next time step in episode")
                    # This should be done elsewhere to prevent locking
                    current_portfolio.step_price(forced=True)
                    break

        if is_done is True:
            break


""" 
    Find an manipulate price information according to how you'd do it inside of an API. 
"""

import random
import threading
import uuid
from copy import copy
from functools import lru_cache
from pprint import pprint
from typing import List

import crayons as cy
import maya
import yfinance as yf
from addict import Dict
from anycache import anycache
from jamboree import Jamboree
from jamboree.handlers import DataHandler, MultiDataManagement
from jamboree.handlers.abstracted.datasets.price import PriceData
from loguru import logger

from see137.search.backtesting import BacktestManager


"""
    Find manipulate price information dynamically like you would from a UI
"""



class UpdateableMultDataManagement(MultiDataManagement):
    def __init__(self):
        super().__init__()
    
    # def update
    


@anycache(cachedir='/tmp/anycache.my')
def get_dataset_info(ds_str: str):
    # ds_str = " ".join(ds_list)
    ticker = yf.Ticker(ds_str)

    return ticker.info


@anycache(cachedir='/tmp/anyticker.my')
def get_ticker_data(ds_str: str):
    data = yf.download(ds_str, group_by='ticker').dropna()
    return data


# Find all datasets by market
class TestSearchTrial(object):
    """ 
        # TEST SEARCH TRIAL

        Class to test running through various scenarios to:

        1. Download data (apple and msft data)
        2. Save it to given metadata inside of redis
        3. Find all various metadata inside of redis using incomplete information
        4. Get the dataset_ids by subtype match
        5. Get datasets by_ids (resemble putting datasets by)
        6. Add datasets into a multidata container
        7. Backtest through the multiple datasets
    """
    def __init__(self):
        # List of datasets we intend to download from yahoo finance
        self.datalist = ["MSFT", "AAPL", "GOOG"]
        self.datasets = []
        self.dataset_info = {}
        self.processor = Jamboree()
        self.pricedata_management = PriceData()
        self.pricedata_management.processor = self.processor
        self.backman = BacktestManager()
        
        self.backman.processor = self.processor
        
        self.dow_metadata = []
        self.reloaded_datasets = []
        self.episodes = []
        self._raw = []
        self.set_name = "Default Test"

        # # pass

    def download_yfinance_info(self):
        for ds in self.datalist:
            info = get_dataset_info(ds)
            self.dataset_info[ds] = {
                "long_name": info["longName"],
            }

    def download_from_yfinance(self):
        self.ds_str = " ".join(self.datalist)
        self.datasets = get_ticker_data(self.ds_str)

        logger.info(
            f"starting download for datasets from yahoo finance {self.ds_str}")
        logger.success(
            f"finished downloading from yahoo finance: {self.ds_str}")

    def save_datasets(self):
        """ Save all of the datasets with the given metadata information. """

        for k, v in self.dataset_info.items():
            abbv = k
            long_name = v['long_name']
            df = self.datasets[k]
            self.pricedata_management.build(name=long_name,
                                            abbv=abbv,
                                            exchange="Dow Jones")
            self.pricedata_management.reset()
            self.pricedata_management.store_time_df(df, is_bar=True)

    def find_all_datasets_by_exchange(self):
        name = "do*"
        dow_jones = self.pricedata_management.search.Find(
            subcategories={"exchange": name})
        self.dow_metadata = dow_jones

        logger.info(f'Finding metadata by partial exchange query \"{name}\"')
        # logger.warning(dow_jones)
        logger.success(f"Successfully loaded metadata ... ")

    def find_meta_by_id(self, _id):
        """ Find a single metadata document by id"""
        return self.pricedata_management.search.pick(_id)

    def from_meta(self, item: dict):
        """ Gets a dataset from metadata """
        dd = Dict(**item)
        entity = dd.metatype
        name = dd.name
        category = dd.category
        subcategories = dd.subcategories
        submetatype = dd.submetatype  # the specific kind of metadata (price, economic)
        abbv = dd.abbreviation

        dataset = DataHandler()
        # dataset.entity = entity
        dataset['name'] = name
        dataset['category'] = category
        dataset['subcategories'] = subcategories
        dataset['submetatype'] = submetatype
        dataset['abbreviation'] = abbv
        dataset.processor = self.processor
        dataset.reset()

        return dataset

    def raw_dict(self, item: dict):
        dd = Dict(**item)
        norm = Dict()
        name = dd.name
        category = dd.category
        subcategories = dd.subcategories
        submetatype = dd.submetatype  # the specific kind of metadata (price, economic)
        abbv = dd.abbreviation
        norm.name = name
        norm.category = category
        norm.subcategories = subcategories
        norm.submetatype = submetatype
        norm.abbreviation = abbv
        return norm.to_dict()

    def get_datasets_by_metadata(self):
        """ Get all of dataset objects by metdata. Used to see if we can """
        data_list: List[DataHandler] = []
        dow_len = len(self.dow_metadata)
        # logger.warning(dow_len)
        if dow_len > 0:
            logger.debug(f"Number of metadata sets: {dow_len}")
            for _ in self.dow_metadata:
                dataset = self.from_meta(_)
                data_list.append(dataset)
                self._raw.append(self.raw_dict(_))

        ll = len(data_list)
        self.reloaded_datasets = data_list
        logger.success(f"Successfully created datasets ... count: {ll}")

    def create_episodes(self):
        """ Create a bunch of episodes """
        self.episodes = [
            uuid.uuid4().hex for x in range(len(self.reloaded_datasets))
        ]
        for index, episode in enumerate(self.episodes):
            dataset = self.reloaded_datasets[index]
            dataset.episode = episode
            dataset.live = False
            dataset.time.head = maya.now().subtract(weeks=600)._epoch
            dataset.time.change_stepsize(microseconds=0, days=1, hours=0)
            dataset.time.change_lookback(microseconds=0, weeks=4, hours=0)
            self.reloaded_datasets[index] = dataset

    def create_multi_data(self):
        self.multi_data = MultiDataManagement()
        self.multi_data["name"] = self.set_name
        self.multi_data["subcategories"] = {"extra": "information"}
        # self.multi_data["submetatype"] = "price"
        self.multi_data["abbreviation"] = "PRACT"
        self.multi_data.processor = self.processor
        self.multi_data.episode = uuid.uuid4().hex
        self.multi_data.description="Default Description"
        self.multi_data.reset()
        self.multi_data.add_multiple_data_sources(self._raw)
        assert len(self.multi_data.sources) > 0, "No sources added to the multi data source"
        # pprint(self.multi_data.sources)

    def run_random_dataset(self):
        dataset = random.choice(self.reloaded_datasets)
        dataset.is_robust = True
        start_time = maya.now()._epoch
        i = 0
        while dataset.is_next:
            # This cuts loop time in half immediately
            dataset.time.step()
            data = dataset.closest_head()
            time_diff = maya.now().epoch - start_time
            i += 1
            milli = (1000 / (i / time_diff))
            logger.info(f"Number of milliseconds: {milli}")
            logger.info(data)

    def run_multidata(self):
        self.multi_data.time.head = maya.now().subtract(weeks=600,
                                                        hours=14)._epoch
        self.multi_data.time.change_stepsize(microseconds=0, days=1, hours=0)
        self.multi_data.time.change_lookback(microseconds=0, weeks=4, hours=0)
        self.multi_data.sync()
        start_time = maya.now()._epoch
        i = 0
        count = len(self.datalist)
        while self.multi_data.is_next:
            self.multi_data.step("current")
            self.multi_data.time.step()
            time_diff = maya.now().epoch - start_time
            i += 1
            milli = (1000 / (i / time_diff))
            time_per_item = (milli / count)
            logger.info(
                f"Number of milliseconds: {time_per_item} for {count} reads totaling - {milli}"
            )

    def run(self, is_download=True):

        if is_download:
            # Downloads the data from yahoo finance to our system
            self.download_yfinance_info()
            self.download_from_yfinance()
            self.save_datasets()

        # Finds every Dow Jones exchanges and stores metadata locally
        self.find_all_datasets_by_exchange()
        # Takes local metadata and finds the datasources in the system
        self.get_datasets_by_metadata()
        # Creates a bunch of backtest episodes (only to iterate through episodes)
        self.create_episodes()
        # Creates a batch dataset (pulling everything at once)
        self.create_multi_data()
        # Get the source ids to locate the universe in search.
        _id = self.multi_data.metaid
        logger.warning(_id)
        latest_multi = self.multi_data.pick(_id)
        asset_ids = latest_multi.source_ids
        logger.debug(asset_ids)


    def run_findselection(self):
        self.find_all_datasets_by_exchange()
        item = random.choice(self.dow_metadata)
        item = Dict(**item)
        refind = self.find_meta_by_id((item.id))
        if refind is None:
            raise AttributeError("item doesn't exist")
        datasource = self.from_meta(refind)
        num = datasource.count()
        last = datasource.closest_head()
        # many = datasource.many()
        logger.warning(num)
        logger.warning(last)
        # logger.warning(many)


if __name__ == "__main__":
    trial = TestSearchTrial()
    trial.run()
    # trial.run_findselection()



class RLEnvironment(object):
    """ Use to setup and start a reinforcement learning training session """
    def __init__(self, *args, **kwargs):
        # Set strategy
        # Set multidataset
        # Set time
        pass
    
    def time(self):
        pass

    def requirements(self):
        """ Creates a requirements handler based on """
        pass


import uuid
from typing import List, Optional

import maya
from cytoolz.curried import filter, get, map, pipe
from cytoolz.itertoolz import unique
from jamboree import Jamboree
from jamboree.base.processors.abstracts import Processor
from jamboree.handlers.abstracted.search import ParameterizedSearch
from jamboree.handlers.default.search import BaseSearchHandler
from see137.search.prototype import SearchFacade
from loguru import logger
from pydantic import PositiveInt, ValidationError, validate_arguments



class _BacktestEpisodeStatus(BaseSearchHandler):
    """ 
        Use to find backtest information by episodes, and episodes by backtest

        - Is there any backtest that's running right now?
        - Is the episode still running?
        - Update the episode's status
        - Start an episode
        - Get the number of episodes that are running
        - What is the strategy we're using for this episode?
        - What are the requirements for this episode?
            - Used so we can pull requirements for the episode 

        ```py
            # Start a new backtest
            backtest_manager.create.episodes.init(backtest_id)

            backtest_manager.update.episodes.status(backtest_id, "...")

            # Get the difference
            backtest_manager.info.episodes.by_backtest(backtest_id)
            backtest_manager.info.episodes.by_episode(episode_id)

            
            # General running episodes
            backtest_manager.info.episodes.total
            backtest_manager.info.episodes.total_active
            backtest_manager.info.episodes.total_inactive



            backtest_manager.info.episodes.active_by(backtest)
            backtest_manager.info.episodes.inactive_by(backtest)
            backtest_manager.info.episodes.is_running(backtest_id, by="backtest")
            backtest_manager.info.episodes.is_running(episode_id, by="episode")

            backtest_manager.info.episodes.status(episode_id, by="episode")
            backtest_manager.info.episodes.status(backtest_id, by="backtest")
        ```
    """
    def __init__(self):
        super().__init__()
        
        self.entity = "episode_backtest_status"
        self.dreq = {"episode": str, "backtest": str, "status":str, 'portfolio':str, "requirements":str, 'strategy':str}
        

class BacktestEpisodeStatus(SearchFacade):
    def __init__(self):
        ProcedureSearch = _BacktestEpisodeStatus()
        super().__init__(ProcedureSearch)
        self.must_have = ['episode', 'backtest']
        self.allowed = ["RUNNING", "DONE"]

    # Universal functions that we'll use in other parts

    def active_by(self, backtest_id:str) -> List[dict]:
        """ Get all active backtests by backtest_id"""
        return self.Find(backtest=backtest_id, status="RUNNING")

    def inactive_by(self, backtest_id:str) -> List[dict]:
        """ Get all inactive backtests by backtest_id"""
        return self.Find(backtest=backtest_id, status="DONE")
    
    def total_active_by(self, backtest_id:str) -> int:
        return len(self.active_by(backtest_id))

    def total_inactive_by(self, backtest_id:str) -> int:
        return len(self.inactive_by(backtest_id))


class EpisodeCreate(BacktestEpisodeStatus):
    """ 
        Use to find backtest information by episodes, and episodes by backtest

        - Is there any backtest that's running right now?
        - Is the episode still running?
        - Update the episode's status
        - Start an episode
        - Get the number of episodes that are running

        ```py
            # Start a new backtest
            backtest_manager.create.episodes.init(backtest_id)

            backtest_manager.update.episodes.status(backtest_id, "...")

            # Get the difference
            backtest_manager.info.episodes.by_backtest(backtest_id)
            backtest_manager.info.episodes.by_episode(episode_id)

            
            # General running episodes
            backtest_manager.info.episodes.total
            backtest_manager.info.episodes.total_active
            backtest_manager.info.episodes.total_inactive



            backtest_manager.info.episodes.active_by(backtest)
            backtest_manager.info.episodes.inactive_by(backtest)
            backtest_manager.info.episodes.is_running(backtest_id, by="backtest")
            backtest_manager.info.episodes.is_running(episode_id, by="episode")

            backtest_manager.info.episodes.status(episode_id, by="episode")
            backtest_manager.info.episodes.status(backtest_id, by="backtest")
        ```
    """
    def __init__(self):
        super().__init__()
    
    def by_backtest(self, backtest_id):
        """ All episodes by backtestid"""
        return self.Find(backtest=backtest_id)

    def init(self, backtest_id:str, **kwargs):
        # logger.bind()
        active_count = self.total_active_by(backtest_id)
        

        
        if active_count == 0:
            episode_id = uuid.uuid4().hex
            _backtest_id = self.Create(no_overwrite_reqs=True, episode=episode_id, backtest=backtest_id, status="RUNNING", **kwargs)
            active_count = self.total_active_by(backtest_id)
            
            return _backtest_id, episode_id, True

        # backtest_id = self.Create(no_overwrite_reqs=True, episode=episode_id, backtest=backtest_id, status="RUNNING", **kwargs)
        
        return backtest_id, None, False
    
    # def init_portfolio()

    
    
class EpisodeUpdate(BacktestEpisodeStatus):
    """ 
        Use to find backtest information by episodes, and episodes by backtest

        - Is there any backtest that's running right now?
        - Is the episode still running?
        - Update the episode's status
        - Start an episode
        - Get the number of episodes that are running

        ```py
            # Start a new backtest
            backtest_manager.create.episodes.init(backtest_id)

            backtest_manager.update.episodes.status(backtest_id, "...")

            # Get the difference
            backtest_manager.info.episodes.by_backtest(backtest_id)
            backtest_manager.info.episodes.by_episode(episode_id)

            
            # General running episodes
            backtest_manager.info.episodes.total
            backtest_manager.info.episodes.total_active
            backtest_manager.info.episodes.total_inactive



            backtest_manager.info.episodes.active_by(backtest)
            backtest_manager.info.episodes.inactive_by(backtest)
            backtest_manager.info.episodes.is_running(backtest_id, by="backtest")
            backtest_manager.info.episodes.is_running(episode_id, by="episode")

            backtest_manager.info.episodes.status(episode_id, by="episode")
            backtest_manager.info.episodes.status(backtest_id, by="backtest")
        ```
    """
    def __init__(self):
        super().__init__()
        
    
    def status(self, episode_id:str, _status="RUNNING"):
        """ Update episode status """
        if _status not in self.allowed:
            raise ValueError("Episode status must be RUNNING or DONE")
        self.UpdateMany({"episode": episode_id}, status=_status)
    
    
    def strategy(self, episode:str, strategy:str):
        self.UpdateMany({"episode": episode}, status="DONE")
    
    def portfolio(self, episode:str, portfolio:str):
        self.UpdateMany({"episode": episode}, portfolio=portfolio)
    
    def requirements(self, episode:str, requirements:str):
        self.UpdateMany({"episode": episode}, requirements=requirements)
    


class EpisodeInfo(BacktestEpisodeStatus):
    """ 
        Use to find backtest information by episodes, and episodes by backtest

        - Is there any backtest that's running right now?
        - Is the episode still running?
        - Update the episode's status
        - Start an episode
        - Get the number of episodes that are running

        ```py
            # Get the difference
            backtest_manager.info.episodes.by_backtest(backtest_id)
            backtest_manager.info.episodes.by_episode(episode_id)

            
            # General running episodes
            backtest_manager.info.episodes.total
            backtest_manager.info.episodes.total_active
            backtest_manager.info.episodes.total_inactive

            # Get active episodes
            # Check if they're running
            backtest_manager.info.episodes.active_by(backtest)
            backtest_manager.info.episodes.inactive_by(backtest)
            backtest_manager.info.episodes.is_running(backtest_id, by="backtest")
            backtest_manager.info.episodes.is_running(episode_id, by="episode")

            backtest_manager.info.episodes.status(episode_id, by="episode")
            backtest_manager.info.episodes.status(backtest_id, by="backtest")
        ```
    """
    def __init__(self):
        super().__init__()
        self.allowed = ["backtest", "episode"]
    
    @property
    def episodes(self) -> List[dict]:
        """ get all episodes """
        universes = self.Find("*")
        unis = pipe(universes, map(lambda uni: uni['episode']), unique, list)
        return unis
    
    @property
    def active(self) -> List[dict]:
        """ Get all active episodes"""
        return self.Find(status="RUNNING")
    
    @property
    def inactive(self) -> List[dict]:
        """ Get all inactive episodes"""
        return self.Find(status="DONE")

    @property
    def total(self) -> int:
        return len(self.episodes)

    @property
    def total_active(self) -> int:
        return len(self.active)
    
    @property
    def total_inactive(self) -> int:
        return len(self.inactive)

    
    def is_running(self, episode_id:str):
        single_episdoe = self.Find(episode=episode_id, status="RUNNING")
        if len(single_episdoe) > 0:
            return True
        return False
        

    def status(self, episode_id:str):
        single_episode = self.Find(episode=episode_id)
        if len(single_episode) > 0:
            return single_episode[0].get("status")
        return "DONE"

    def by_backtest(self, backtest_id):
        """ All episodes by backtestid"""
        return self.Find(backtest=backtest_id)
    
    def by_episode(self, episode_id:str):
        """ All episodes by backtestid"""
        _episode = self.Find(episode=episode_id)
        if len(_episode) > 0:
            return _episode[0]
        return {}

class EpisodeRemove(BacktestEpisodeStatus):
    """ 
        
    """
    def __init__(self):
        super().__init__()

    def by_id(self, episode_id:str):
        """ Remove episode by id"""
        self.Remove(episode=episode_id)

    def by_backtest(self, backtest_id:str):
        self.Remove(backtest=backtest_id)
    
    


def manage_episodes():
    proc = Jamboree()
    episode_create = EpisodeCreate()
    episode_update = EpisodeUpdate()
    episode_info = EpisodeInfo()
    episode_remove = EpisodeRemove()


    episode_create.processor = proc
    episode_update.processor = proc
    episode_info.processor = proc
    episode_remove.processor = proc

    backtest_id = "3e667a9ade4c4a409ae153522f46cc3d"
    # episode_remove.general = "*"
    # episode_remove.remove()
    episode_remove.RemoveAll()
    backtest_id, episode_id, is_running = episode_create.init(backtest_id)
    print(episode_info.total_active)
    print(episode_info.status(episode_id))
    print(episode_info.is_running(episode_id))
    episode_update.status(episode_update, _status="DONE")
    print(episode_info.status(episode_id))
    print(episode_info.total_active)
    print(episode_info.total_inactive)
    print(episode_info.total_active_by(backtest_id))
    episode_remove.by_id(episode_id)
    print(episode_info.total_inactive)

if __name__ == "__main__":
    manage_episodes()



from typing import Any
from abc import ABC, abstractmethod
from see137.common.injectable import InjectableObject

class ComponentAbstract(ABC, InjectableObject):
    
    @abstractmethod
    def set_settings(self, **kwargs):
        raise NotImplementedError("No way to manage dynamic component settings.")

    @abstractmethod
    def step(self, item:Any, **kwargs):
        raise NotImplementedError("Execute hasn't been implemented")


import copy
import random
import threading
from queue import Queue
import uuid
from concurrent.futures import ThreadPoolExecutor
from contextlib import ContextDecorator
from typing import Any, Dict, List, Optional

import maya
import numpy as np
import pandas as pd
# from crayons import blue, cyan, green, magenta, red, white, yellow
from toolz import keyfilter
from jamboree import DBHandler
from jamboree import Jamboree
from jamboree.handlers.default import MultiDataManagement, TimeHandler
from jamboree.handlers.default.data import DataHandler
from jamboree.utils import omit
from loguru import logger

from see137.data import PriceGenerator
# from linktt_end.handlers.allocation import AllocationHandler
from see137.handlers.allocation import AllocationHandler
from see137.utils.slippage import RandomUniformSlippageModel
from see137.utils.trades.trade import DynamicTrade, Trade, TradeType
#
# I need Arunavo to finish order management.
#
logger.disable(__name__)


"""
Note: Will remove soon. Going through a concept.
"""


class PortfolioHandler(DBHandler):
    """Abstract handler that we use to keep track of portfolio information.
    """

    def __init__(self,
                 base_instrument="USD",
                 start_balance=10000,
                 limit=100,
                 lazy_load=False):
        super().__init__()
        self.entity = "portfolio"
        self.required = {
            "episode": str,
            "user_id": str,
            "exchange": str,
            "live": bool
        }

        # TODO: Lazy load should be replaced with the parameters `pre_load` and `just_in_time`
        self._base_instrument = base_instrument
        self.start_balance = start_balance
        # TODO: Cut down on the object initiation.
        # NOTE: We can initiate the object once then change the rest of the variables multiple times
        self._time: TimeHandler = TimeHandler()
        self._data: MultiDataManagement = MultiDataManagement()
        self._allocator: Optional[AllocationHandler] = AllocationHandler()

        self._balance = 0
        self.limit = limit
        self._subcategories = {}

        # Store a list of the portfolio & performance states

        self._slippage_model = RandomUniformSlippageModel()
        # TODO: Need to replace
        self._instrument_precision = 8
        self._commission_percent = 0.3
        self._base_precision = 2

        self._min_trade_amount = 0.000001
        self._max_trade_amount = 1000000
        self._min_trade_price = 0.00000001
        self._max_trade_price = 100000000

        # Placeholder variables
        self.placeholders = {}
        self.latest_prices = {}
        # Determines if the object has loaded the important variables. Use to prevent loading again
        self.has_updated = False
        self.is_inner_allocation = False
        self.is_sim = False

        # TODO: Figure out the settings at some point

    # -------------------------------------------------------------------
    # --------------------- Properties & Setters ------------------------
    # -------------------------------------------------------------------
    @property
    def is_live(self) -> bool:
        lively = self['live']
        episode = self['episode']
        if episode == "live" and lively == True:
            return True
        return False

    @property
    def current_time(self):
        """ 
            # Current Time
            ---
            Get the current time for the portfolio.

            Gets the time from the price/dataset. If the datasource is live it should get the current price and use that. 
        """
        # TODO: Add live logic
        # If episode and live are both live, return maya.now()._epoch
        return self.time.head

    @property
    def limit(self):
        return self._limit

    @limit.setter
    def limit(self, limit):
        self._limit = limit

    @property
    def user_id(self) -> str:
        return self['user_id']

    @user_id.setter
    def user_id(self, _user_id:str):
        self['user_id'] = _user_id

    @property
    def exchange(self) -> str:
        return self['exchange']
    
    @exchange.setter
    def exchange(self, _exchange:str):
        self['exchange'] = _exchange


    @property
    def episode(self) -> str:
        return self['episode']
    
    @episode.setter
    def episode(self, _episode:str):
        self['episode'] = _episode
        

    @property
    def holdings(self):
        """ Get the holdings for the user """
        if self.has_updated:
            return self.placeholders.get("holdings", [])
        _holdings = self.load_holdings()
        filtration_list = [
            'user_id', 'episode', 'type', 'detail', 'exchange', 'live',
            'timestamp', 'event_id'
        ]
        holdings = map(lambda x: omit(filtration_list, x), _holdings)
        return list(holdings)

    @property
    def performance(self):
        """ Get the performance of the user/exchange"""
        if self.has_updated == True:
            return self.placeholders.get("performance", {})
        performance = self.load_performance()
        if isinstance(performance, dict):
            performance = [performance]

        filtration_list = [
            'user_id', 'episode', 'type', 'timestamp', 'exchange', 'live',
            'session', 'time', 'event_id'
        ]

        performance_filtered = list(
            map(lambda x: omit(filtration_list, x), performance))
        return performance_filtered

    @property
    def portfolio(self):
        """ Get the latest portfolio of the user/exchange"""
        if self.has_updated == True:
            return self.placeholders.get("portfolio", {})
        _portfolio = self.latest_portfolio()
        filtered_portfolio = omit([
            'user_id', 'episode', 'type', 'timestamp', 'exchange', 'live',
            'session', 'time', 'event_id', 'detail'
        ], _portfolio)
        return filtered_portfolio

    @property
    def portfolio_history(self):
        """ Get the latest portfolio of the user/exchange"""
        if self.has_updated:
            return self.placeholders.get("portfolio_history", [])
        portfolio = self.load_portfolio()
        filtration_list = [
            'user_id', 'episode', 'type', 'timestamp', 'exchange', 'live',
            'event_id', 'detail'
        ]
        filtered_portfolio = list(
            map(lambda x: omit(filtration_list, x), portfolio))
        return filtered_portfolio

    @property
    def balance(self) -> float:
        if self.has_updated:
            return self.placeholders.get("balance", 0.0)
        latest_balance = self.latest_performance()
        return float(latest_balance.get('balance', 0.0))

    @property
    def trades(self) -> List[Dict]:
        """ Get all of the trades for this user"""
        trades = self.load_trades()
        if isinstance(trades, dict):
            trades = [trades]
        filtration_list = [
            "episode", "exchange", "user_id", "type", "detail", "timestamp",
            "live", 'event_id'
        ]
        trades_filtered = map(lambda x: omit(filtration_list, x), trades)
        return list(trades_filtered)

    @property
    def net_worth(self) -> float:
        net_worth = self.balance
        portfolio = self.portfolio
        if not portfolio:
            return net_worth
        portfolio_filtered = omit([
            'user_id', 'episode', 'type', 'time', 'timestamp', 'exchange',
            'live', 'event_id', 'detail', 'mtype'
        ], portfolio)
        for symbol, amount in portfolio_filtered.items():
            if symbol == self._base_instrument:
                continue

            current_price = self.current_price(symbol=symbol)
            net_worth += current_price * amount

        return float(net_worth)

    @property
    def pricing(self):
        """ 
            NOTE:
                - We determine the user's portfolio using this dataset.
                - Since it has a name heavily tailored to the
                - We position what we want the user to have using this pricing indicator 
        """
        if self._data is None:
            raise NotImplementedError("Pricing handler not found")
        user_id = self['user_id']
        exchange = self['exchange']
        episode = self['episode']
        live = self['live']

        self._data['name'] = f"{user_id}-{exchange}-{episode}-{live}"
        self._data['category'] = "portfolio_info"
        self._data['subcategories'] = {"real": False}
        self._data['metatype'] = "port"
        self._data['submetatype'] = "lio"
        self._data['abbreviation'] = "UPORT"

        self._data.episode = episode
        self._data.live = live
        self._data.processor = self.processor
        # self._data.reset()
        return self._data

    def is_user_asset(self, asset: dict) -> bool:
        """ Determines if the given asset is indeed the user's portfolio """
        asset_keys = list(asset.keys())
        # logger.error(asset_keys)
        if len(asset_keys) == 0:
            return False

        # TODO: Fix the source of this
        if 'episode' in asset_keys:
            episode = asset.pop('episode', None)
            asset['subcategories']['episode'] = episode
        # logger.error(asset_keys)
        current_user_assets = self.users_assets
        # logger.info(current_user_assets[0])
        # logger.success(asset)
        if len(current_user_assets) == 0:
            return False
        hashed_comparision_asset = self.main_helper.generate_hash(asset)
        # logger.info(hashed_comparision_asset)
        comparison_hashed_assets = [
            self.main_helper.generate_hash(x) for x in current_user_assets
        ]
        return hashed_comparision_asset in comparison_hashed_assets

    @property
    def users_assets(self) -> list:
        return self.pricing.sources

    @property
    def allocation(self) -> dict:
        if self.has_updated:
            return self.placeholders.get("allocation", {})
        allocator = self.allocator.allocation
        return allocator

    @property
    def allocator(self):
        """ Allocation stuff for"""
        if self._allocator is None:
            raise AttributeError("Allocator handler not found")

        user_id = self['user_id']
        exchange = self['exchange']
        episode = self['episode']
        live = self['live']

        self._allocator["episode"] = episode
        self._allocator["user_id"] = user_id
        self._allocator["exchange"] = exchange
        self._allocator["live"] = live
        self._allocator.processor = self.processor
        self._allocator.time = self.time
        return self._allocator

    @property
    def profit_loss_percent(self) -> float:
        """
            Calculate the percentage change in net worth since the last reset.
            Returns:
                The percentage change in net worth since the last reset.
        """
        return float(self.net_worth / self.start_balance) * 100

    # Use to get counts inside of the database

    @property
    def time(self) -> 'TimeHandler':
        self._time.processor = self.processor
        self._time["live"] = self["live"]
        self._time["episode"] = self["episode"]
        return self._time

    def get_count(self) -> int:
        alt = {"detail": "portfolio"}
        count = self.count(alt=alt)
        return count

    def performance_count(self) -> int:
        alt = {"detail": "performance"}
        count = self.count(alt)
        return count

    def holdings_count(self) -> int:
        alt = {"detail": "holdings"}
        count = self.count(alt)
        return count

    def _get_mock_backpressure(self, start:float, stop:float, backpressure_prob:float=0.1, detail_name:str="amount"):
        mock_name = f"{detail_name}_mock"
        alt = {"detail": mock_name}
        latest_value = self.get_single(alt=alt)
        is_backpressure = (random.uniform(0, 1) < backpressure_prob)
        if not bool(latest_value) or not is_backpressure:
            value = random.uniform(0, 30)
            val_dict = {"value": value}
            self.set_single(val_dict, alt=alt)
            return value
        return latest_value['value']
        

    def current_amount(self, symbol: str, is_mocked=False, backpressure_prob:float=0.1) -> float:
        """Get the amount (count) of a given asset.

        Args:
            symbol (str): The name of the asset we're looking for.
            is_mocked (bool, optional): Used to set a mock. Returns a random variable if it's set to true. Defaults to False.

        Returns:
            float: The amount of the assets in the portfolio.
        """
        if is_mocked is True: return self._get_mock_backpressure(0, 30, backpressure_prob=backpressure_prob, detail_name="amount")
        self.load_placeholders()
        portfolio = self.portfolio
        # how many of this given asset do we have
        amount_value = portfolio.get(symbol, 0.000000000000000000001)
        return float(amount_value)

    def current_allocation(self, symbol="BTC", is_mocked=False, backpressure_prob:float=0.1) -> float:
        """Get the current allocation of a given asset.

        Args:
            symbol (str, optional): The symbol we're looking to get allocations for. Defaults to "BTC".

        Returns:
            float: The allocation of the symbol. Defaults to 0.0 if we can't find it.
        """
        if is_mocked is True: return self._get_mock_backpressure(0, 0.999, backpressure_prob=backpressure_prob, detail_name="allocation")
        allocation = self.allocation
        # how many of this given asset do we have
        allocation_value = allocation.get(symbol, 0.000000000000000000001)
        return float(allocation_value)
    
    def current_pct_worth(self, symbol:str="BTC", is_mocked:bool=True, backpressure_prob:float=0.1) -> float:
        """current_pct_worth Percentage Of Networth

        Get the percentage of the networth for the given symbol.

        Args:
            symbol (str, optional): The symbol we're checking. Defaults to "BTC".
            is_mocked (bool, optional): Determines if we're mocking the results. Defaults to True.

        Returns:
            float: Percentage of networth the asset takes.
        """
        net_worth = self.net_worth
        current_value = self.current_total(symbol=symbol, is_mocked=is_mocked, backpressure_prob=backpressure_prob)
        return current_value/net_worth

    def current_pct_limit(self, symbol="BTC", is_mocked:bool=False, backpressure_prob:float=0.1) -> float:
        """current_pct_limit Get the percentage of limit.

        Given the portfolio allocation cap that we've set. The allocation cap is the maximum amount we're allowed to trade on within a given time period.

        Args:
            symbol (str, optional): Symbol to get info for. Defaults to "BTC".
            is_mocked (bool, optional): Determines if we're mocking. Defaults to False.

        Returns:
            float: The percentage of the maximum allocation cap.
        """
        pct_worth:float = self.current_pct_worth(symbol=symbol, is_mocked=is_mocked, backpressure_prob=backpressure_prob)
        allocation_pct:float = self.current_allocation(symbol=symbol, is_mocked=is_mocked, backpressure_prob=backpressure_prob)
        if allocation_pct == 0: return 0.0
        return pct_worth/allocation_pct

    def current_price(self, symbol:str="BTC", is_mocked:bool=False, backpressure_prob:float=0.1) -> float:
        """current_price Get the price of `symbol`.

        Returns the current price of the given asset.

        if `is_mocked` is `True` we return a random number between 0 and 1000. 

        Args:
            symbol (str, optional): The symbol we're getting the price for. Defaults to "BTC".
            is_mocked (bool, optional): Determines if we're mocking the price. Defaults to False.


        Returns:
            float: The current price.
        """
        if is_mocked is True: return self._get_mock_backpressure(0, 1000, backpressure_prob=backpressure_prob, detail_name="price")
        
        current_price = self.latest_prices.get(symbol, 0.000000000000000000001)
        return float(current_price)
    

    def current_total(self, symbol:str="BTC", is_mocked:bool=False, backpressure_prob:float=0.1) -> float:
        """# Total Value
        
        Gets the total value of the given symbol.

        ```
            total = price * amount
        ```
        
        Args:
            symbol (str, optional): The symbol of the asset we're looking to get price information for. Defaults to "BTC".
            is_mocked (bool, optional): Determines if we're mocking the price. Defaults to False.

        Returns:
            float: The total value of the asset given in the symbol.
        """
        price = self.current_price(symbol=symbol, is_mocked=is_mocked, backpressure_prob=backpressure_prob)
        amount = self.current_amount(symbol=symbol, is_mocked=is_mocked, backpressure_prob=backpressure_prob)
        return price * amount

    # ----------------------------------------
    # -------------- Querying ----------------
    # ----------------------------------------

    def latest_portfolio(self) -> Dict:
        """Load the latest portfolio state.

        Returns:
            Dict: The portfolio state. All of the assets we're currently holding.
        """
        alt = {"detail": "portfolio"}
        if self.is_live == True:
            last_state = self.last(alt=alt, ar="absolute")
            return last_state
        last_state = self.last(alt=alt, ar="relative")
        return last_state

    def latest_performance(self):
        alt = {"detail": "performance"}
        if self.is_live == True:
            last_state = self.last(alt=alt, ar="absolute")
            return last_state
        last_state = self.last(alt=alt, ar="relative")
        return last_state

    def load_performance(self):
        alt = {"detail": "performance"}

        if self.is_live == True:
            performance = self.many(self.limit, alt=alt, ar="absolute")
        else:
            performance = self.many(self.limit, alt=alt, ar="relative")

        if isinstance(performance, dict):
            return [performance]
        return performance

    def load_portfolio(self):
        alt = {"detail": "portfolio"}
        if self.is_live == True:
            portfolio = self.many(self.limit, alt=alt, ar="absolute")
        else:
            portfolio = self.many(self.limit, alt=alt, ar="relative")
        if isinstance(portfolio, dict):
            return [portfolio]
        return portfolio

    def load_trades(self):
        alt = {"detail": "trade"}
        if self.is_live == True:
            trades = self.many(self.limit, alt=alt, ar="absolute")
        else:
            trades = self.many(self.limit, alt=alt, ar="relative")
        if isinstance(trades, dict):
            return [trades]
        return trades

    def load_holdings(self):
        alt = {"detail": "holdings"}
        if self.is_live == True:
            _holdings = self.many(self.limit, alt=alt, ar="absolute")
        else:
            _holdings = self.many(self.limit, alt=alt, ar="relative")
        if isinstance(_holdings, dict):
            return [_holdings]
        return _holdings

    def load_placeholders(self):
        if self.has_updated == False:
            # pass
            self.placeholders['portfolio'] = self.portfolio
            self.placeholders['balance'] = float(self.balance)
            self.placeholders['holdings'] = self.holdings
            self.placeholders['portfolio_history'] = self.portfolio_history
            self.placeholders['performance'] = self.performance
            self.placeholders['allocation'] = self.allocation
            self.has_updated = True
            # TODO: Set the allocation here

    # ----------------------------------------
    # ---------------- Saving ----------------
    # ----------------------------------------

    def save_portfolio(self, data):
        """ Save portfolio """
        alt = {"detail": "portfolio"}
        _data = copy.copy(self._query)
        _data.update(data)
        _data['type'] = self.entity
        _data['time'] = self.current_time
        _data['timestamp'] = maya.now()._epoch
        for key, value in _data.items():
            if isinstance(value, np.float64):
                _data[key] = float(value)
        self.save(_data, alt=alt)

    def save_performance(self, data) -> None:
        """ Save performance information """
        alt = {"detail": "performance"}
        _data = copy.copy(self._query)
        _data.update(data)
        _data.update(alt)
        _data['type'] = self.entity
        _data['time'] = self.current_time
        _data['timestamp'] = maya.now()._epoch
        for key, value in _data.items():
            if isinstance(value, np.float64):
                _data[key] = float(value)
        self.save(_data, alt=alt)

    def save_trade(self, data):
        alt = {"detail": "trade"}

        _data = copy.copy(self._query)
        _data.update(data)
        _data.update(alt)

        _data['type'] = self.entity
        _data['time'] = self.current_time
        _data['timestamp'] = maya.now()._epoch
        for key, value in _data.items():
            if isinstance(value, np.float64):
                _data[key] = float(value)
        self.save(_data, alt=alt)

    def save_holdings(self, data, timestamp=maya.now()._epoch):
        alt = {"detail": "holdings"}

        _data = copy.copy(self._query)
        _data.update(data)
        _data.update(alt)

        _data['type'] = self.entity
        _data['time'] = self.current_time
        _data['timestamp'] = maya.now()._epoch

        for key, value in _data.items():
            if isinstance(value, np.float64):
                _data[key] = float(value)

        self.save(_data, alt=alt)

    # ----------------------------------------
    # ------------ Operations ----------------
    # ----------------------------------------

    def _update_account(self, _trade: Trade):
        """ Updates the portfolio exchange account. """
        # TODO: Add timestamp to trade

        if self._is_valid_trade(_trade):
            self._make_trade(_trade)

        current_balance = float(self.balance)
        portfolio = self.portfolio
        portfolio[self._base_instrument] = current_balance

        latest_performance = {
            'net_worth': self.net_worth,
            'balance': current_balance
        }

        self.save_performance(latest_performance)
        self.save_portfolio(portfolio)
        self._update_holdings()

    def _update_account_no_order(self):
        """ Updates the portfolio exchange account. """
        # TODO: Add timestamp to trade
        balance = float(self.balance)
        current_balance = float(balance)
        portfolio = self.portfolio
        portfolio[self._base_instrument] = current_balance

        latest_performance = {
            'net_worth': self.net_worth,
            'balance': current_balance
        }

        self.save_performance(latest_performance)
        self.save_portfolio(portfolio)
        self._update_holdings()

    def _update_holdings(self):
        """ Duplication of net_worth code"""
        holdings = {}
        portfolio = self.portfolio
        if not portfolio:
            return

        portfolio_filtered = omit([
            'user_id', 'episode', 'type', 'time', 'timestamp', 'exchange',
            'live', 'event_id', 'detail', 'mtype'
        ], portfolio)
        

        portfolio_items = portfolio_filtered.items()
        for symbol, amount in portfolio_items:
            if symbol == self._base_instrument:
                holdings[symbol] = amount
                continue

            current_price = self.current_price(symbol=symbol)
            total = current_price * amount
            holdings[symbol] = total

        self.save_holdings(holdings, timestamp=self.current_time)

    def _make_trade(self, trade: Trade):
        """ Trades on the account then updates it"""

        if not trade.is_hold:
            """ TODO: We'll need to save this in two places to make it work IRL"""
            self.save_trade({
                'symbol': trade.symbol,
                'trade_type': trade.trade_type.value,
                'amount': trade.amount,
                'price': trade.price
            })

        if trade.is_buy:
            self.placeholders['balance'] -= float((trade.amount * trade.price))
            self.placeholders['portfolio'][
                trade.symbol] = self.placeholders['portfolio'].get(
                    trade.symbol, 0) + trade.amount
        elif trade.is_sell:
            self.placeholders['balance'] += float((trade.amount * trade.price))
            self.placeholders['portfolio'][
                trade.symbol] = self.placeholders['portfolio'].get(
                    trade.symbol, 0) - trade.amount

    def _is_valid_trade(self, _trade: Trade) -> bool:

        if _trade.is_buy and self.placeholders[
                'balance'] < _trade.amount * _trade.price:
            logger.error("Not enough to buy")
            return False
        elif _trade.is_sell and self.placeholders['portfolio'].get(
                _trade.symbol, 0) < _trade.amount:
            logger.error("Not enough to sell")
            return False

        is_both = _trade.amount >= self._min_trade_amount and _trade.amount <= self._max_trade_amount

        # if is_both: logger.success("Both conditions succeeded")
        return is_both

    def execute_trade(self, trade: Trade) -> Trade:
        self.load_placeholders()

        current_price = self.current_price(symbol=trade.symbol)
        commission = self._commission_percent / 100
        filled_trade = trade.copy()

        filled_price = filled_trade.price
        if filled_price == 0:
            filled_price = 0.000000001

        if filled_trade.is_hold or not self._is_valid_trade(filled_trade):
            filled_trade.amount = 0

        if filled_trade.is_buy:
            price_adjustment = (1 + commission)
            filled_trade.price = round(current_price * price_adjustment,
                                       self._base_precision)
            filled_trade.amount = round(
                (filled_trade.price * filled_trade.amount) / filled_price,
                self._instrument_precision)
        elif filled_trade.is_sell:
            price_adjustment = (1 - commission)
            filled_trade.price = round(current_price * price_adjustment,
                                       self._base_precision)
            filled_trade.amount = round(filled_trade.amount,
                                        self._instrument_precision)

        if not filled_trade.is_hold:
            filled_trade = self._slippage_model.fill_order(
                filled_trade, current_price)

        self._update_account(filled_trade)
        return filled_trade

    def instrument_balance(self, symbol: str):
        """ Get the balance for the instrument """
        portfolio = self.latest_portfolio()

        if symbol in portfolio.keys():
            return portfolio[symbol]
        return 0.0

    # -------------------------------------------------------
    # ------------------ Reset Conditions -------------------
    # -------------------------------------------------------

    def _reset_portfolio(self):
        count = self.get_count()
        if count == 0:
            self.save_portfolio({
                f"{self._base_instrument}":
                float(self.start_balance),
                "time":
                self.current_time
            })

    def _reset_performance(self):
        """ Reset the performance inside of a given exchange """
        count = self.performance_count()
        if count == 0:
            self.save_performance({
                "balance": float(self.start_balance),
                "net_worth": float(self.start_balance),
                "time": self.current_time
            })

    def _reset_holdings(self):
        """ Reset the holdings inside of a given exchange """
        count = self.holdings_count()
        if count == 0:
            self.save_holdings({
                f"{self._base_instrument}":
                float(self.start_balance),
                "time":
                self.current_time
            })

    def _reset_price(self):
        """ Resets the price. If it's a backtest or simulation, we'll pull information into place so we can backtest. """
        self.pricing.reset()

    def _reset_allocation(self):
        self.allocator.reset()

    def _reset_time(self):
        self.time.reset()

    def add_asset(self,
                  name,
                  subcategories={},
                  extras={
                      "submetatype": "",
                      'abbreviation': ""
                  }):
        """This can only be under the category market"""
        category = "markets"
        combined_dict = {
            "name": name,
            "category": category,
            "subcategories": subcategories
        }
        combined_dict.update(extras)
        if self.is_sim:
            combined_dict["subcategories"]['episode'] = self['episode']
        self.pricing.add_multiple_data_sources([combined_dict])

    def remove_asset(self, name, subcategories={}):
        """
            TODO: Add remove asset command to the class during the hardening.
            Remove assets from the user. 
            It doesn't add if it doesn't exist.
        """
        pass

    def price_preprocessing(self, data: dict):
        data_keys = list(data.keys())
        first_item = [x.split(":")[0] for x in data_keys]
        portfolio = pd.DataFrame()
        latest_price_dict = {}
        for index, first in enumerate(first_item):
            dk = data_keys[index]
            d1 = data[dk]
            portfolio[str(first)] = d1['close']
            latest_price_dict[first] = d1['close'].iloc[-1]
        return portfolio, latest_price_dict

    def price_preprocessing_current(self, data: Dict[str, Dict[Any, Any]]) -> Dict[str, Any]:
        """ Get all of the information and return the latest price dict"""
        latest_price_dict = {}
        for key, value in data.items():
            first = key.split(":")[0]
            close = value.get("close", 0.0)
            latest_price_dict[first] = close
        return latest_price_dict

    def clear_cache(self):
        self.placeholders = {}
        self.has_updated = False
    
    def clear_update(self):
        self.clear_cache()
        self.load_placeholders()

    def sim_only_data(self, datasets):
        if self.is_sim == False:
            return

        if len(datasets) == 0:
            return
        subcategory = datasets[0]['subcategories']
        subcategory['episode'] = self['episode']
        new_list = []
        for dataset in datasets:
            dataset['subcategories'] = subcategory
            new_list.append(dataset)
        self._data.data_handler_list = new_list

    def step_price(self, forced: bool = False) -> None:
        self.pricing.sync()

        if self.allocator.should_rebalance and forced:
            prices_information          = self.pricing.step()
            portfolio_batch, price_dict = self.price_preprocessing(prices_information)
            self.latest_prices          = price_dict
            with logger.catch():
                self.allocator.step(portfolio_batch)
            self.has_updated = False
            self.load_placeholders()
            return

        prices_information = self.pricing.step(call_type="current")
        price_dict = self.price_preprocessing_current(prices_information)

        self.latest_prices = price_dict
        self.allocator._finish_step()

    def step_trade(self, trade: Trade = None) -> None:
        if trade is not None:
            if isinstance(trade, DynamicTrade):
                net_worth = self.net_worth

                price  : float = self.current_price(symbol=trade.symbol)
                amount : float = self.current_amount(symbol=trade.symbol)
                cap_pct: float = self.current_allocation(symbol=trade.symbol)

                dynamic_trade: DynamicTrade = copy.copy(trade)
                dynamic_trade.cap_pct       = cap_pct

                dynamic_trade.amount    = amount
                dynamic_trade.price     = price
                dynamic_trade.net_worth = net_worth

                trade = dynamic_trade.calculate_trade()

                if isinstance(trade.amount, np.float64):
                    trade.amount = float(trade.amount)

            if isinstance(trade, Trade):
                logger.success(trade)
                self.execute_trade(trade)

            else:
                self._update_account_no_order()
        else:
            self._update_account_no_order()

    def step(self, trade: Optional[Trade] = None) -> Dict:
        final_dict = {'obs': {}, 'done': False}
        """ It's a step sandwich"""
        self.step_price()
        # This doesn't load if we've stepped through and reallocated the class internally
        self.load_placeholders()
        self.step_trade(trade)

        final_dict['done'] = (not self.pricing.is_next)
        # we'd put all of the portfolio information here by asset.
        final_dict['obs']["portfolio"] = {}
        # get percentages
        self.clear_cache()
        return final_dict

    def render(self):
        """ Get the information necessary to give a report for the user. """
        self.load_placeholders()

        performance_frame:pd.DataFrame = pd.DataFrame(self.performance)
        holdings_frame   :pd.DataFrame = pd.DataFrame(self.holdings)
        portfolio_frame  :pd.DataFrame = pd.DataFrame(self.portfolio_history)

        print(performance_frame.tail(3))
        print(holdings_frame.tail(3))
        print(portfolio_frame.tail(3))

    def reset(self):
        """ Determines if we're re-initiating """
        logger.warning("Resetting Portfolio")
        self.check()
        self._reset_price()
        self._reset_time()
        self._reset_allocation()

        """ Portfolio specific resets """
        self._reset_performance()
        self._reset_portfolio()
        self._reset_holdings()


def get_base_datahandler():
    datahandler = DataHandler()
    datahandler["category"] = "markets"
    datahandler["subcategories"] = {
        "market": "stock",
        "country": "US",
        "sector": "faaaaake"
    }
    return datahandler


def price_gen_func(episodes: List[str],
                   asset_list=["BTC", "ATL", "TRX", "ETH", "BCH", "XRP"],
                   number_of_episodes=10):
    if episodes is None:
        episodes = [uuid.uuid4().hex for x in range(number_of_episodes)]
    price_gen = PriceGenerator()
    price_gen.episodes = episodes
    price_gen.assets = asset_list
    price_gen.starting_min = 20
    price_gen.starting_max = 5000
    price_gen.is_varied = True
    price_gen.length = 3000
    price_gen.generate_all_episodes()
    return price_gen


def main():
    """ Should run through everything returning zero right now"""
    asset_list = ["BTC", "ATL", "TRX", "ETH", "BCH", "XRP"]

    main_episode = uuid.uuid4().hex
    user_id = uuid.uuid4().hex
    exchange_name = "fake_exchange"
    episodes = [main_episode]
    prices = price_gen_func(episodes=episodes,
                            asset_list=asset_list).episode_set

    processor = Jamboree()

    portfolio_hand = PortfolioHandler()
    portfolio_hand.processor = processor
    portfolio_hand['exchange'] = exchange_name
    portfolio_hand['user_id'] = user_id
    portfolio_hand['episode'] = main_episode
    portfolio_hand['live'] = False
    portfolio_hand.reset()
    portfolio_hand.time.head = maya.now().add(weeks=25)._epoch
    portfolio_hand.time.change_stepsize(microseconds=0, days=1, hours=0)
    portfolio_hand.time.change_lookback(microseconds=0, weeks=25, hours=0)

    dh = get_base_datahandler()
    dh.processor = processor
    dh.reset()
    subcat = dh['subcategories']
    portfolio_hand.is_sim = True
    for episode in prices.keys():
        logger.warning(f"Episode: {episode}")
        for asset in asset_list:
            _subcat = copy.copy(subcat)
            _subcat['episode'] = episode
            dh["name"] = asset
            dh['subcategories'] = _subcat
            data = prices[episode][asset]
            dh.reset()
            dh.store_time_df(data, is_bar=True)

    portfolio_hand.add_asset(asset_list[0], subcat)
    portfolio_hand.add_asset(asset_list[1], subcat)
    portfolio_hand.add_asset(asset_list[2], subcat)
    portfolio_hand.is_inner_allocation = True
    portfolio_hand.step_price()
    for _ in range(100):
        if random.uniform(0, 1) < 0.1:
            symbol = random.choice(
                [asset_list[0], asset_list[1], asset_list[2]])
            trade_type = random.choice(
                [TradeType.LIMIT_BUY, TradeType.LIMIT_SELL])
            dynamic_trade = DynamicTrade(symbol, trade_type)
            dynamic_trade.percentage = random.normalvariate(0.3, 0.06)
            portfolio_hand.step(dynamic_trade)
        else:
            portfolio_hand.step()
        portfolio_hand.time.step()
    portfolio_hand.render()


if __name__ == "__main__":
    main()



import time
import uuid
from datetime import datetime

import maya
from faker import Faker
from pydantic import BaseModel, Field

fake = Faker()


def random_day():
    start = fake.date_between(start_date="-10y")
    current_datetime = datetime.fromordinal(start.toordinal())
    mayatime = maya.MayaDT.from_datetime(current_datetime)
    return mayatime._epoch


def hex_uuid():
    return uuid.uuid4().hex


class BacktestConfig(BaseModel):
    name:str = Field(default_factory=fake.company)
    description: str = Field(default_factory=fake.sentence)
    user:str = Field(default_factory=hex_uuid)

    base:str = "USD"
    initial:str = 10000.0
    exchange:str = "binance"

    start: float = Field(default_factory=random_day)
    end: float = Field(default_factory=random_day)
    lookback: dict = {}
    step: dict = {}

    strategy:str = Field(default_factory=hex_uuid)
    universe:str = Field(default_factory=hex_uuid)




import uuid
from typing import Dict, Any
from addict import Dict as ADict
from see137.common import ComponentAbstract, AddictWithManagerComponent


class DecideStrategyComponent(AddictWithManagerComponent):
    """Decide Strategy

        Pull strategy from the database and make decision
    """
    def __init__(self):
        super().__init__()
        self.name = "Decide Strategy"
        # Backtest configuration dict. Can be saved into the database immediately.
        self.required_items:Dict[str, type] = {"strategy_blob":object, "episode": str, "data" : object}

        # Make sure the data is already pulled (We'll probably be casting this into a dataframe)
        # self.required_settings: Dict[str, type] = {}
    
    def process(self, item:ADict):
        strategy_blob = item.strategy_blob
        # Check strategy type (leave out for now)
        preprocessed_data = strategy_blob.preprocess(item)
        action, percentage = strategy_blob.step(preprocessed_data)
        item.action = action
        item.percentage = percentage

    def step(self, item:ADict, **kwargs):
        return super().__init__(item, **kwargs)


import uuid
import maya

from typing import Optional, List
from jamboree import Jamboree
from jamboree.base.processors.abstracts import Processor
from jamboree.handlers.abstracted.search import ParameterizedSearch
from jamboree.handlers.default.search import BaseSearchHandler
from see137.search.prototype import SearchFacade
from loguru import logger


class EpisodeSpecificInformationSearch(BaseSearchHandler):
    """
        A search handler for all episode specific information. Use to find related information per episode.

        This is a variation of session specific information, which we'll use to get reward function by episode.

        
    """
    def __init__(self):
        super().__init__()

        self.entity = "episodespec"
        self.dreq = {
            'episode': str,
            'backtest': str,
            'strategy': str,
            'portfolio': str
        }


class BacktestEpisodeSpec(SearchFacade):
    def __init__(self):
        ProcedureSearch = EpisodeSpecificInformationSearch()
        super().__init__(ProcedureSearch)
        self.must_have = ['episode']

    def strategy_by_episode(self, episode: str):
        strategies = self.Find(episode=episode)
        if len(strategies) > 0:
            strategy_dict = strategies[0]
            strategy_id = strategy_dict['strategy']
            return strategy_id
        raise AttributeError("Strategy Doesn't Exist")


from .handlers.exchange import PortfolioHandler
from .handlers.allocation import AllocationHandler
from .handlers.requirements import RequirementsHandler
from .common.creation import SimulationCreator
from .common.accessor import SingleAccessor


class InjectableObject:
    _processor = None

    @property
    def processor(self):
        if self._processor is None:
            raise AttributeError("Injected processor hasn't been set. ")
        return self._processor

    @processor.setter
    def processor(self, _processor):
        self._processor = _processor
    
    def set_processor(self, _processor):
        self._processor = _processor


import maya
import uuid
from typing import Union, Optional
# import see137.parameters.fields i
from see137.parameters.fields import TimeRange, MovementSpecification
from jamboree.handlers.default import TimeHandler
from jamboree.utils.core import consistent_hash
from see137.parameters.parameter import Parameter
from loguru import logger

class TimeParameter(Parameter):
    def __init__(self, interval:TimeRange, steps:MovementSpecification, lookback:MovementSpecification, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.interval = interval
        self.stepinfo = steps
        self.lbinfo = lookback
        if not isinstance(self.interval, TimeRange):
            raise ValueError("Interval needs to be a time range")
        if not isinstance(self.stepinfo, MovementSpecification):
            raise ValueError("Steps configuration data needs to be a RangeSpecificField.")
        if not isinstance(self.lbinfo, MovementSpecification):
            raise ValueError("Lookback parameters supplied needs to be RangeSpecificField.")
    
    @property
    def start(self) -> float:
        """Return the starting epoch

        Returns:
            float -- the epoch for the start of the backtest. Set the head to this
        """
        return self.interval.start.epoch

    @property
    def end(self) -> float:
        """Return the ending epoch

        Returns:
            float -- the epoch for the end of the backtest. Check for this at the end of steps for multiple assets. 
        """
        return self.interval.end.epoch
    
    @property
    def step(self) -> str:
        """Return Step Size Information

        Returns:
            str -- string describing how much we're going to be moving the observation needle ahead.
        """
        return self.stepinfo.to_json()
    
    @property
    def lookback(self) -> str:
        """Return Lookback Size Information

        Returns:
            str -- string describing how much we're going to be looking backwards during the backtest.
        """
        return self.lbinfo.to_json()
    
    @property
    def duration(self):
        return self.interval.interval.duration
    

@logger.catch
def main():
    start = "2007-05-02"
    end = "2020-04-22"
    timezone = "UTC"
    interval = TimeRange(start=start, end=end, timezone=timezone)
    steps = MovementSpecification(days=2)
    lookback = MovementSpecification(weeks=2)

    

    time_conf = TimeParameter(interval=interval, steps=steps, lookback=lookback)
    print(time_conf.start)
    print(time_conf.end)
    print(time_conf.lookback)
    print(time_conf.step)
    print(time_conf.duration)


if __name__ == "__main__":
    main()


import numpy
from see137.stochastic import is_acceptable_type
from see137.stochastic.utils.parameters import get_super_parameters
from see137.stochastic.generators import (
    geometric_brownian_motion_levels,
    heston_model_levels,
    geometric_brownian_motion_jump_diffusion_levels,
    cox_ingersoll_ross_levels,
    ornstein_uhlenbeck_levels
)

def generate_super_price(base_price:float=1000.0, base_type:str="MERTON", base_time:int=2000) -> numpy.array:


    if is_acceptable_type(base_type) == False:
        raise ValueError("Stochastic type is not appropiately specified")

    super_params = get_super_parameters(base_price, base_time)
    
    if base_type.upper() == "GBM":
        m = geometric_brownian_motion_levels(super_params)
        return m
    elif base_type.upper() == "HESTON":
        m = heston_model_levels(super_params)[0]
        return m
    elif base_type.upper() == "MERTON":
        m = geometric_brownian_motion_jump_diffusion_levels(super_params)
        return m
    elif base_type.upper() == "COX":
        m = cox_ingersoll_ross_levels(super_params)
        return m
    elif base_type.upper() == "ORNSTEIN":
        m = ornstein_uhlenbeck_levels(super_params)
        return m


from loguru import logger
from addict import Dict as Addict
from typing import List, Any, Union, Optional
from see137.common import Pipeline
from see137.common import ComponentAbstract


class AddictPipeline(Pipeline):
    def __init__(self,
                 steps: List[Union['Pipeline', 'ComponentAbstract']] = [],
                 name: str = None,
                 processor=None):
        super().__init__(steps, name=name, processor=processor)

    def step(self,
             dict_item: dict,
             override: bool = False,
             is_loud: bool = True,
             **kwargs):
        addict_dict = Addict(**dict_item)
        
        kwargs.update({"override": override, "is_loud": is_loud})
        result = super().step(addict_dict, **kwargs)
        return result
    
    def __repr__(self):
        return f"<Pipeline={self.name}>"


# As a user, I'd like to pull the attached process strategy, pull the data, process the data, get a decision from that data
from typing import Dict, List

from addict import Dict as ADict
from jamboree import Jamboree
from jamboree.handlers.default.multi import MultiDataManagement
from loguru import logger

from see137.common import (AddictWithManagerComponent,
                                    ComponentAbstract, Pipeline, PipelineA)
from see137.components.strategy import CheckStrategy
from see137.components.universe import CheckUniverse
from see137.handlers.strategy.core import StrategyEngine
from see137.models.backtests_models import BacktestConfig

class CreateBacktestComponent(AddictWithManagerComponent):
    """Save Backtest Component

        Pull strategy from the database to do work on.
    """
    def __init__(self):
        super(CreateBacktestComponent, self).__init__()


        # Backtest configuration dict. Can be saved into the database immediately.
        self.required_items: Dict[str, type] = {
            "config": dict, "valid_multi":bool, 'valid_strategy':bool
        }

    def check_prev(self, item):
        return super().check_prev(item)

    def check_all(self, item):
        return super().check_all(item)
    
    def check_item(self, item):
        return super().check_item(item)

    def raise_loud(self, exc=Exception, msg="General Problem"):
        super().raise_loud(exc=exc, msg=msg)

    def add_backtest_info(self, item: ADict, _id: str, is_save: bool = False):
        item.backtest_id = _id
        item.is_saved = False
        return item

    def process(self, item: ADict):
        """Take the backtest_config and save it into the database.

        Arguments:
            item {ADict} -- Addict modified dictionary. Save's according to backtest data.

        Returns:
            ADict -- Modified dictionary with action and percentage inside of it
        """

        backtestid = self.manager.create.backtest.init(**item.config.to_dict())
        item.backtestid = backtestid
        logger.error(backtestid)
        return item
